# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

import os
import pathlib
import subprocess
import sys
from typing import List


class CommandRunner:
    def __init__(
        self,
        binary_path: pathlib.Path,
        list_arguments: List[str],
        emplacement: pathlib.Path,
    ) -> None:
        self.current_dir: pathlib.Path = pathlib.Path().cwd()
        self.command: pathlib.Path = binary_path
        self.emplacement: pathlib.Path = emplacement
        self.arguments: List[str] = list_arguments

    def check_command(self) -> bool:
        if not self.command.is_file():
            print(f"{self.command} executable not found")
            return False
        return True

    def run(self) -> int:
        if not self.check_command():
            # TODO For now, it will return 0 as if nothing is wrong (not to have errors in CI as executable will not be there)
            # eventually if should return an error
            # maybe wait when we separate unit tests from integration tests
            # modify with bender_decomposed's read_solution
            print("Return code 0 for now")
            return 0

        os.chdir(self.emplacement)
        res = subprocess.run(
            [self.current_dir / self.command, *self.arguments],
            stdout=sys.stdout,
            stderr=subprocess.DEVNULL,  # TODO For now, to avoid the "Invalid MIT-MAGIC-COOKIE-1 key" error
            shell=False,
        )
        os.chdir(self.current_dir)

        return res.returncode


class BendersRunner(CommandRunner):
    def __init__(self, emplacement: pathlib.Path) -> None:
        super().__init__(pathlib.Path("bin/benders"), ["options.json"], emplacement)


class MergeMPSRunner(CommandRunner):
    def __init__(self, emplacement: pathlib.Path) -> None:
        super().__init__(pathlib.Path("bin/merge_mps"), ["options.json"], emplacement)



# output_values_base.py

import math
from abc import ABC, abstractmethod
from dataclasses import dataclass, field
from typing import Dict, List, Optional, Tuple, Union, cast

from gems.study.data import TimeScenarioIndex


@dataclass
class BaseOutputValue(ABC):
    """
    Abstract Base Class providing common structure and functionality
    for Variable and ExtraOutput classes. It handles storage (_value, _size)
    and shared logic (is_close, value property).
    """

    _name: str
    _value: Dict[TimeScenarioIndex, float] = field(init=False, default_factory=dict)
    _size: Tuple[int, int] = field(init=False, default=(0, 0))
    ignore: bool = field(default=False, init=False)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BaseOutputValue):
            return NotImplemented
        return (self.ignore or other.ignore) or (
            self._name == other._name
            and self._size == other._size
            and self._value == other._value
        )

    def is_close(
        self,
        other: "BaseOutputValue",
        *,
        rel_tol: float = 1.0e-9,
        abs_tol: float = 0.0,
    ) -> bool:
        if not isinstance(other, BaseOutputValue):
            return NotImplemented

        return (self.ignore or other.ignore) or (
            self._name == other._name
            and self._size == other._size
            and self._value.keys() == other._value.keys()
            and all(
                math.isclose(
                    self._value[key],
                    other._value[key],
                    rel_tol=rel_tol,
                    abs_tol=abs_tol,
                )
                for key in self._value
            )
        )

    def __str__(self) -> str:
        return f"{self._name} : {str(self.value)} {'(ignored)' if self.ignore else ''}"

    @property
    def value(self) -> Union[None, float, List[float], List[List[float]]]:
        size_s, size_t = self._size
        if size_t == 1:
            if size_s == 1:
                return self._value.get(TimeScenarioIndex(0, 0))
            else:
                return [self._value[TimeScenarioIndex(0, s)] for s in range(size_s)]
        else:
            return [
                [self._value[TimeScenarioIndex(t, s)] for t in range(size_t)]
                for s in range(size_s)
            ]

    @value.setter
    def value(self, values: Union[float, List[float], List[List[float]]]) -> None:
        size_s, size_t = 1, 1
        self._value.clear()
        if isinstance(values, list):
            size_s = len(values)
            for scenario, timesteps in enumerate(values):
                if isinstance(timesteps, list):
                    size_t = len(timesteps)
                    for timestep, value in enumerate(timesteps):
                        self._value[TimeScenarioIndex(timestep, scenario)] = value
                else:
                    self._value[TimeScenarioIndex(0, scenario)] = cast(float, timesteps)
        else:
            self._value[TimeScenarioIndex(0, 0)] = values
        self._size = (size_s, size_t)

    def get(self, t: int, s: int) -> float | None:
        return self._value.get(TimeScenarioIndex(t, s))



from typing import Optional, List
from ..modified_base_model import ModifiedBaseModel
from .gems_component_parameter import GemsComponentParameter


class GemsComponent(ModifiedBaseModel):
    id: str
    model: str
    scenario_group: Optional[str] = None
    parameters: Optional[List[GemsComponentParameter]] = None



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.
import logging
from math import inf
from pathlib import Path

import pandas as pd
from pypsa import Network

from gems.pypsa_converter.utils import any_to_float
from gems.pypsa_converter.models.pypsa_model_schema import (
    PyPSAComponentData,
    PyPSAGlobalConstraintData,
)

from gems.pypsa_converter.models.gems_system_yaml_schema import (
    GemsComponent,
    GemsComponentParameter,
    GemsPortConnection,
    GemsSystem,
)


class PyPSAStudyConverter:
    def __init__(
        self,
        pypsa_network: Network,
        logger: logging.Logger,
        system_dir: Path,
        series_dir: Path,
        series_file_format: str,
    ):
        """
        Initialize processor
        """
        self.logger = logger
        self.system_dir = system_dir
        self.series_dir = series_dir
        self.pypsa_network = pypsa_network
        self.pypsalib_id = "pypsa_models"
        self.null_carrier_id = "null"
        self.system_name = pypsa_network.name
        self.series_file_format = series_file_format

        self.pypsa_components = [
            "buses",
            "loads",
            "generators",
            "stores",
            "storage_units",
            "links",
            "lines",
            "transformers",
        ]
        self._pypsa_network_assertion()
        self._pypsa_network_preprocessing()
        self._preprocess_pypsa_components("loads", False, "/")
        self._preprocess_pypsa_components("generators", True, "p_nom")
        self._preprocess_pypsa_components("stores", True, "e_nom")
        self._preprocess_pypsa_components("storage_units", True, "p_nom")
        self._preprocess_pypsa_components("links", True, "p_nom")

        self.pypsa_components_data: dict[str, PyPSAComponentData] = {}
        self._register_pypsa_components()
        self.pypsa_globalconstraints_data: dict[str, PyPSAGlobalConstraintData] = {}
        self._register_pypsa_globalconstraints()

    def _pypsa_network_assertion(self) -> None:
        """Assertion function to keep trace of the limitations of the converter"""
        assert len(self.pypsa_network.investment_periods) == 0
        assert (self.pypsa_network.snapshot_weightings.values == 1.0).all()
        ### PyPSA components : Generators
        if not (all((self.pypsa_network.generators["marginal_cost_quadratic"] == 0))):
            raise ValueError(f"Converter supports only Generators with linear cost")
        if not (all((self.pypsa_network.generators["active"] == 1))):
            raise ValueError(f"Converter supports only Generators with active = 1")
        if not (all((self.pypsa_network.generators["committable"] == False))):
            raise ValueError(
                f"Converter supports only Generators with commitable = False"
            )
        ### PyPSA components : Loads
        if not (all((self.pypsa_network.loads["active"] == 1))):
            raise ValueError(f"Converter supports only Loads with active = 1")
        ### PyPSA components : Links
        if not (all((self.pypsa_network.links["active"] == 1))):
            raise ValueError(f"Converter supports only Links with active = 1")
        ### PyPSA components : Lines
        if not len(self.pypsa_network.lines) == 0:
            raise ValueError(f"Converter does not support Lines yet")
        ### PyPSA components : Storage Units
        if not (all((self.pypsa_network.links["active"] == 1))):
            raise ValueError(f"Converter supports only Storage Units with active = 1")
        if not (all((self.pypsa_network.storage_units["sign"] == 1))):
            raise ValueError(f"Converter supports only Storage Units with sign = 1")
        if not (all((self.pypsa_network.storage_units["cyclic_state_of_charge"] == 1))):
            raise ValueError(
                f"Converter supports only Storage Units with cyclic_state_of_charge"
            )
        if not (
            all((self.pypsa_network.storage_units["marginal_cost_quadratic"] == 0))
        ):
            raise ValueError(f"Converter supports only Storage Units with linear cost")
        ### PyPSA components : Stores
        if not (all((self.pypsa_network.links["active"] == 1))):
            raise ValueError(f"Converter supports only Stores with active = 1")
        if not (all((self.pypsa_network.stores["sign"] == 1))):
            raise ValueError(f"Converter supports only Stores with sign = 1")
        if not (all((self.pypsa_network.stores["e_cyclic"] == 1))):
            raise ValueError(f"Converter supports only Stores with e_cyclic = True")
        if not (all((self.pypsa_network.stores["marginal_cost_quadratic"] == 0))):
            raise ValueError(f"Converter supports only Stores with linear cost")
        ### PyPSA components : GlobalConstraint
        for pypsa_model_id in self.pypsa_network.global_constraints.index:
            assert (
                self.pypsa_network.global_constraints.loc[pypsa_model_id, "type"]
                == "primary_energy"
            )
            assert (
                self.pypsa_network.global_constraints.loc[
                    pypsa_model_id, "carrier_attribute"
                ]
                == "co2_emissions"
            )

    def _rename_buses(self) -> None:
        ### Rename PyPSA buses, to delete spaces
        if len(self.pypsa_network.buses) > 0:
            self.pypsa_network.buses.index = self.pypsa_network.buses.index.str.replace(
                " ", "_"
            )
            for _, val in self.pypsa_network.buses_t.items():
                val.columns = val.columns.str.replace(" ", "_")
        ### Update the 'bus' columns for the different types of PyPSA components
        for component_type in self.pypsa_components:
            df = getattr(self.pypsa_network, component_type)
            if len(df) > 0:
                for col in ["bus", "bus0", "bus1"]:
                    if col in df.columns:
                        df[col] = df[col].str.replace(" ", "_")

    def _pypsa_network_preprocessing(self) -> None:
        ###Add fictitious carrier
        self.pypsa_network.add(
            "Carrier",
            self.null_carrier_id,
            co2_emissions=0,
            max_growth=any_to_float(inf),
        )
        self.pypsa_network.carriers[
            "carrier"
        ] = self.pypsa_network.carriers.index.values
        self._rename_buses()

    def _rename_pypsa_components(self, component_type: str) -> None:
        df = getattr(self.pypsa_network, component_type)
        if len(df) == 0:
            return
        ### Rename PyPSA components, to make sure that the names are uniques (used as id in the Gems model)
        prefix = component_type[:-1]
        df.index = prefix + "_" + df.index.str.replace(" ", "_")
        dictionnary = getattr(self.pypsa_network, component_type + "_t")
        for _, val in dictionnary.items():
            val.columns = prefix + "_" + val.columns.str.replace(" ", "_")

    def _fix_capacities(self, component_type: str, capa_str: str) -> None:
        df = getattr(self.pypsa_network, component_type)
        if len(df) == 0:
            return
        ### Adding min and max capacities to non-extendable objects
        for field in [capa_str + "_min", capa_str + "_max"]:
            df.loc[df[capa_str + "_extendable"] == False, field] = df[capa_str]
            df.loc[df[capa_str + "_extendable"] == False, "capital_cost"] = 0.0

    def _preprocess_pypsa_components(
        self, component_type: str, extendable: bool, capa_str: str
    ) -> None:
        ### Handling PyPSA objects without carriers
        df = getattr(self.pypsa_network, component_type)
        for comp in df.index:
            if len(df.loc[comp, "carrier"]) == 0:
                df.loc[comp, "carrier"] = self.null_carrier_id
        setattr(
            self.pypsa_network,
            component_type,
            df.join(
                self.pypsa_network.carriers,
                on="carrier",
                how="left",
                rsuffix="_carrier",
            ),
        )
        self._rename_pypsa_components(component_type)
        if extendable:
            self._fix_capacities(component_type, capa_str)

    def _register_pypsa_components(self) -> None:
        ### PyPSA components : Generators
        self._register_pypsa_components_of_given_model(
            "generators",
            self.pypsa_network.generators,
            self.pypsa_network.generators_t,
            "generator",
            {
                "p_nom_min": "p_nom_min",
                "p_nom_max": "p_nom_max",
                "p_min_pu": "p_min_pu",
                "p_max_pu": "p_max_pu",
                "marginal_cost": "marginal_cost",
                "capital_cost": "capital_cost",
                "e_sum_min": "e_sum_min",
                "e_sum_max": "e_sum_max",
                "sign": "sign",
                "efficiency": "efficiency",
                "co2_emissions": "emission_factor",
            },
            {"bus": ("p_balance_port", "p_balance_port")},
        )
        ### PyPSA components : Loads
        self._register_pypsa_components_of_given_model(
            "loads",
            self.pypsa_network.loads,
            self.pypsa_network.loads_t,
            "load",
            {
                "p_set": "p_set",
                "q_set": "q_set",
                "sign": "sign",
            },
            {"bus": ("p_balance_port", "p_balance_port")},
        )
        ### PyPSA components : Buses
        self._register_pypsa_components_of_given_model(
            "buses",
            self.pypsa_network.buses,
            self.pypsa_network.buses_t,
            "bus",
            {
                "v_nom": "v_nom",
                "x": "x",
                "y": "y",
                "v_mag_pu_set": "v_mag_pu_set",
                "v_mag_pu_min": "v_mag_pu_min",
                "v_mag_pu_max": "v_mag_pu_max",
            },
            {},
        )
        ### PyPSA components : Links
        self._register_pypsa_components_of_given_model(
            "links",
            self.pypsa_network.links,
            self.pypsa_network.links_t,
            "link",
            {
                "efficiency": "efficiency",
                "p_nom_min": "p_nom_min",
                "p_nom_max": "p_nom_max",
                "p_min_pu": "p_min_pu",
                "p_max_pu": "p_max_pu",
                "marginal_cost": "marginal_cost",
                "capital_cost": "capital_cost",
            },
            {
                "bus0": ("p0_port", "p_balance_port"),
                "bus1": ("p1_port", "p_balance_port"),
            },
        )
        ### PyPSA components : Storage Units
        self._register_pypsa_components_of_given_model(
            "storage_units",
            self.pypsa_network.storage_units,
            self.pypsa_network.storage_units_t,
            "storage_unit",
            {
                "p_nom_min": "p_nom_min",
                "p_nom_max": "p_nom_max",
                "p_min_pu": "p_min_pu",
                "p_max_pu": "p_max_pu",
                "sign": "sign",
                "efficiency_store": "efficiency_store",
                "efficiency_dispatch": "efficiency_dispatch",
                "standing_loss": "standing_loss",
                "max_hours": "max_hours",
                "marginal_cost": "marginal_cost",
                "capital_cost": "capital_cost",
                "marginal_cost_storage": "marginal_cost_storage",
                "spill_cost": "spill_cost",
                "inflow": "inflow",
                "co2_emissions": "emission_factor",
            },
            {"bus": ("p_balance_port", "p_balance_port")},
        )
        ### PyPSA components : Stores
        self._register_pypsa_components_of_given_model(
            "stores",
            self.pypsa_network.stores,
            self.pypsa_network.stores_t,
            "store",
            {
                "sign": "sign",
                "e_nom_min": "e_nom_min",
                "e_nom_max": "e_nom_max",
                "e_min_pu": "e_min_pu",
                "e_max_pu": "e_max_pu",
                "standing_loss": "standing_loss",
                "marginal_cost": "marginal_cost",
                "capital_cost": "capital_cost",
                "marginal_cost_storage": "marginal_cost_storage",
                "co2_emissions": "emission_factor",
            },
            {"bus": ("p_balance_port", "p_balance_port")},
        )

    def _add_contributors_to_globalconstraints(
        self, gems_components_and_ports: list[tuple[str, str]], component_type: str
    ) -> list[tuple[str, str]]:
        df = getattr(self.pypsa_network, component_type)
        gems_components_and_ports += [
            (comp, "emission_port")
            for comp in df[df["carrier"] != self.null_carrier_id].index
        ]
        return gems_components_and_ports

    def _register_pypsa_globalconstraints(self) -> None:
        gems_components_and_ports: list[tuple[str, str]] = []
        for component_type in ["generators", "stores", "storage_units"]:
            gems_components_and_ports = self._add_contributors_to_globalconstraints(
                gems_components_and_ports, component_type
            )

        for pypsa_model_id in self.pypsa_network.global_constraints.index:
            name, sense, carrier_attribute = (
                pypsa_model_id,
                self.pypsa_network.global_constraints.loc[pypsa_model_id, "sense"],
                self.pypsa_network.global_constraints.loc[
                    pypsa_model_id, "carrier_attribute"
                ],
            )
            if carrier_attribute == "co2_emissions" and sense == "<=":
                self.pypsa_globalconstraints_data[
                    pypsa_model_id
                ] = PyPSAGlobalConstraintData(
                    name,
                    carrier_attribute,
                    sense,
                    self.pypsa_network.global_constraints.loc[
                        pypsa_model_id, "constant"
                    ],
                    "global_constraint_co2_max",
                    "emission_port",
                    gems_components_and_ports,
                )
            elif carrier_attribute == "co2_emissions" and sense == "==":
                self.pypsa_globalconstraints_data[
                    pypsa_model_id
                ] = PyPSAGlobalConstraintData(
                    name,
                    carrier_attribute,
                    sense,
                    self.pypsa_network.global_constraints.loc[
                        pypsa_model_id, "constant"
                    ],
                    "global_constraint_co2_eq",
                    "emission_port",
                    gems_components_and_ports,
                )
            else:
                raise ValueError("Type of GlobalConstraint not supported.")

    def _register_pypsa_components_of_given_model(
        self,
        pypsa_model_id: str,
        constant_data: pd.DataFrame,
        time_dependent_data: dict[str, pd.DataFrame],
        gems_model_id: str,
        pypsa_params_to_gems_params: dict[str, str],
        pypsa_params_to_gems_connections: dict[str, tuple[str, str]],
    ) -> None:
        if pypsa_model_id in self.pypsa_components_data:
            raise ValueError(f"{pypsa_model_id} already registered !")

        self.pypsa_components_data[pypsa_model_id] = PyPSAComponentData(
            pypsa_model_id,
            constant_data,
            time_dependent_data,
            gems_model_id,
            pypsa_params_to_gems_params,
            pypsa_params_to_gems_connections,
        )

    def to_gems_study(self) -> GemsSystem:
        """Main function, to export PyPSA as Gems system"""

        self.logger.info("Study conversion started")
        list_components, list_connections = [], []

        for pypsa_components_data in self.pypsa_components_data.values():
            components, connections = self._convert_pypsa_components_of_given_model(
                pypsa_components_data
            )
            list_components.extend(components)
            list_connections.extend(connections)

        for pypsa_global_constraint_data in self.pypsa_globalconstraints_data.values():
            (
                components,
                connections,
            ) = self._convert_pypsa_globalconstraint_of_given_model(
                pypsa_global_constraint_data
            )
            list_components.extend(components)
            list_connections.extend(connections)

        return GemsSystem(
            nodes=[], components=list_components, connections=list_connections
        )

    def _convert_pypsa_components_of_given_model(
        self, pypsa_components_data: PyPSAComponentData
    ) -> tuple[list[GemsComponent], list[GemsPortConnection]]:
        """
        Generic function to handle the different PyPSA classes

        """

        self.logger.info(
            f"Creating objects of type: {pypsa_components_data.gems_model_id}. "
        )

        # We test whether the keys of the conversion dictionary are allowed in the PyPSA model : all authorized parameters are columns in the constant data frame (even though they are specified as time-varying values in the time-varying data frame)
        pypsa_components_data.check_params_consistency()

        # List of params that may be time-dependent in the pypsa model, among those we want to keep
        time_dependent_params = set(
            pypsa_components_data.pypsa_params_to_gems_params
        ).intersection(set(pypsa_components_data.time_dependent_data.keys()))
        # Save time series and memorize the time-dependent parameters
        comp_param_to_timeseries_name = self._write_and_register_timeseries(
            pypsa_components_data.time_dependent_data, time_dependent_params
        )

        connections = self._create_gems_connections(
            pypsa_components_data.constant_data,
            pypsa_components_data.pypsa_params_to_gems_connections,
        )

        components = self._create_gems_components(
            pypsa_components_data.constant_data,
            pypsa_components_data.gems_model_id,
            pypsa_components_data.pypsa_params_to_gems_params,
            comp_param_to_timeseries_name,
        )
        return components, connections

    def _convert_pypsa_globalconstraint_of_given_model(
        self, pypsa_gc_data: PyPSAGlobalConstraintData
    ) -> tuple[list[GemsComponent], list[GemsPortConnection]]:
        self.logger.info(
            f"Creating PyPSA GlobalConstraint of type: {pypsa_gc_data.gems_model_id}. "
        )
        components = [
            GemsComponent(
                id=pypsa_gc_data.pypsa_name,
                model=f"{self.pypsalib_id}.{pypsa_gc_data.gems_model_id}",
                parameters=[
                    GemsComponentParameter(
                        id="quota",
                        time_dependent=False,
                        scenario_dependent=False,
                        value=pypsa_gc_data.pypsa_constant,
                    )
                ],
            )
        ]
        connections = []
        for component_id, port_id in pypsa_gc_data.gems_components_and_ports:
            connections.append(
                GemsPortConnection(
                    component1=pypsa_gc_data.pypsa_name,
                    port1=pypsa_gc_data.gems_port_id,
                    component2=component_id,
                    port2=port_id,
                )
            )

        return components, connections

    def _write_and_register_timeseries(
        self,
        time_dependent_data: dict[str, pd.DataFrame],
        time_dependent_params: set[str],
    ) -> dict[tuple[str, str], str]:
        comp_param_to_timeseries_name = dict()
        for param in time_dependent_params:
            param_df = time_dependent_data[param]
            for component in param_df.columns:
                timeseries_name = self.system_name + "_" + component + "_" + param
                comp_param_to_timeseries_name[(component, param)] = timeseries_name
                param_df[[component]].to_csv(
                    self.series_dir / Path(timeseries_name + self.series_file_format),
                    index=False,
                    header=False,
                )

        return comp_param_to_timeseries_name

    def _create_gems_components(
        self,
        constant_data: pd.DataFrame,
        gems_model_id: str,
        pypsa_params_to_gems_params: dict[str, str],
        comp_param_to_timeseries_name: dict[tuple[str, str], str],
    ) -> list[GemsComponent]:
        components = []
        for component in constant_data.index:
            components.append(
                GemsComponent(
                    id=component,
                    model=f"{self.pypsalib_id}.{gems_model_id}",
                    parameters=[
                        GemsComponentParameter(
                            id=pypsa_params_to_gems_params[param],
                            time_dependent=(component, param)
                            in comp_param_to_timeseries_name,
                            scenario_dependent=False,
                            value=(
                                comp_param_to_timeseries_name[(component, param)]
                                if (component, param) in comp_param_to_timeseries_name
                                else any_to_float(constant_data.loc[component, param])
                            ),
                        )
                        for param in pypsa_params_to_gems_params
                    ],
                )
            )
        return components

    def _create_gems_connections(
        self,
        constant_data: pd.DataFrame,
        pypsa_params_to_gems_connections: dict[str, tuple[str, str]],
    ) -> list[GemsPortConnection]:
        connections = []
        for bus_id, (
            model_port,
            bus_port,
        ) in pypsa_params_to_gems_connections.items():
            buses = constant_data[bus_id].values
            for component_id, component in enumerate(constant_data.index):
                connections.append(
                    GemsPortConnection(
                        component1=buses[component_id],
                        port1=bus_port,
                        component2=component,
                        port2=model_port,
                    )
                )
        return connections



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from dataclasses import dataclass


@dataclass(frozen=True)
class IndexingStructure:
    """
    Specifies if parameters and variables should be indexed by time and/or scenarios.
    """

    time: bool
    scenario: bool

    def __or__(self, other: "IndexingStructure") -> "IndexingStructure":
        time = self.time or other.time
        scenario = self.scenario or other.scenario
        return IndexingStructure(time, scenario)

    def is_time_varying(self) -> bool:
        return self.time

    def is_scenario_varying(self) -> bool:
        return self.scenario

    def is_time_scenario_varying(self) -> bool:
        return self.is_time_varying() and self.is_scenario_varying()

    def is_constant(self) -> bool:
        return (not self.is_time_varying()) and (not self.is_scenario_varying())



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from .copy import CopyVisitor, copy_expression
from .degree import ExpressionDegreeVisitor, compute_degree
from .evaluate import EvaluationContext, EvaluationVisitor, ValueProvider, evaluate
from .evaluate_parameters import (
    ParameterResolver,
    ParameterValueProvider,
    resolve_parameters,
)
from .expression import (
    AdditionNode,
    Comparator,
    ComparisonNode,
    DivisionNode,
    ExpressionNode,
    LiteralNode,
    MultiplicationNode,
    NegationNode,
    ParameterNode,
    VariableNode,
    literal,
    param,
    sum_expressions,
    var,
)
from .print import PrinterVisitor, print_expr
from .visitor import ExpressionVisitor, visit



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

import os
from pathlib import Path, PurePath
from typing import Any, Union

import pandas as pd
import yaml
from pandas import DataFrame
from pydantic import BaseModel

from gems.input_converter.src.parsing import ConversionTemplate


def resolve_path(path: Path) -> Path:
    if not path.exists():
        raise FileNotFoundError

    absolute_path = path.resolve()
    return absolute_path


def check_file_exists(input_path: Path) -> bool:
    if input_path.exists() and input_path.is_file() and input_path.stat().st_size > 0:
        return True
    return False


def check_dataframe_validity(df: DataFrame) -> bool:
    """
    Check and validate the following conditions:
    1. The dataframe from this path is not empty.
    2. The dataframe does not contains only zero values.

    :param df: dataframe to validate.
    """
    if df.empty or (df == 0).all().all():
        return False

    return True


def dump_to_yaml(model: BaseModel, output_path: Path) -> None:
    with open(output_path, "w", encoding="utf-8") as yaml_file:
        yaml.dump(
            {
                "system": model.model_dump(by_alias=True, exclude_unset=True),
            },
            yaml_file,
            allow_unicode=True,
        )


def read_yaml_file(file_path: Path) -> dict[str, Any]:
    if not file_path.exists():
        raise FileNotFoundError(f"The file {file_path} does not exists")
    with file_path.open("r", encoding="utf-8") as file:
        try:
            return yaml.safe_load(file)
        except yaml.YAMLError as e:
            raise yaml.YAMLError(f"Error trying to read yaml file {file_path}: {e}")


def save_to_file(series: Union[pd.DataFrame, pd.Series], output_file: PurePath) -> None:
    output_dir = os.path.dirname(output_file)
    if output_dir:
        os.makedirs(output_dir, exist_ok=True)

    series.to_csv(output_file, sep="\t", index=False, header=False)



# Generated from Expr.g4 by ANTLR 4.13.2
from antlr4 import *

if "." in __name__:
    from .ExprParser import ExprParser
else:
    from ExprParser import ExprParser

# This class defines a complete generic visitor for a parse tree produced by ExprParser.


class ExprVisitor(ParseTreeVisitor):
    # Visit a parse tree produced by ExprParser#portFieldExpr.
    def visitPortFieldExpr(self, ctx: ExprParser.PortFieldExprContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#fullexpr.
    def visitFullexpr(self, ctx: ExprParser.FullexprContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#portFieldSum.
    def visitPortFieldSum(self, ctx: ExprParser.PortFieldSumContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#negation.
    def visitNegation(self, ctx: ExprParser.NegationContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#unsignedAtom.
    def visitUnsignedAtom(self, ctx: ExprParser.UnsignedAtomContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#expression.
    def visitExpression(self, ctx: ExprParser.ExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#comparison.
    def visitComparison(self, ctx: ExprParser.ComparisonContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#allTimeSum.
    def visitAllTimeSum(self, ctx: ExprParser.AllTimeSumContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#timeIndexExpr.
    def visitTimeIndexExpr(self, ctx: ExprParser.TimeIndexExprContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#addsub.
    def visitAddsub(self, ctx: ExprParser.AddsubContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#timeShiftExpr.
    def visitTimeShiftExpr(self, ctx: ExprParser.TimeShiftExprContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#portField.
    def visitPortField(self, ctx: ExprParser.PortFieldContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#muldiv.
    def visitMuldiv(self, ctx: ExprParser.MuldivContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#timeSum.
    def visitTimeSum(self, ctx: ExprParser.TimeSumContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#timeIndex.
    def visitTimeIndex(self, ctx: ExprParser.TimeIndexContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#timeShift.
    def visitTimeShift(self, ctx: ExprParser.TimeShiftContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#function.
    def visitFunction(self, ctx: ExprParser.FunctionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#number.
    def visitNumber(self, ctx: ExprParser.NumberContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#identifier.
    def visitIdentifier(self, ctx: ExprParser.IdentifierContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#shift.
    def visitShift(self, ctx: ExprParser.ShiftContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#signedAtom.
    def visitSignedAtom(self, ctx: ExprParser.SignedAtomContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#signedExpression.
    def visitSignedExpression(self, ctx: ExprParser.SignedExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#shiftMuldiv.
    def visitShiftMuldiv(self, ctx: ExprParser.ShiftMuldivContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#shiftAddsub.
    def visitShiftAddsub(self, ctx: ExprParser.ShiftAddsubContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#rightExpression.
    def visitRightExpression(self, ctx: ExprParser.RightExpressionContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#rightMuldiv.
    def visitRightMuldiv(self, ctx: ExprParser.RightMuldivContext):
        return self.visitChildren(ctx)

    # Visit a parse tree produced by ExprParser#rightAtom.
    def visitRightAtom(self, ctx: ExprParser.RightAtomContext):
        return self.visitChildren(ctx)


del ExprParser



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

"""
Utility classes to obtain solver results.
"""
import math
from dataclasses import dataclass, field
from typing import Any, Dict, Mapping, Optional, Set, TypeVar

from gems.expression import evaluate
from gems.expression.evaluate import EvaluationError
from gems.simulation.extra_output import ExtraOutput, ExtraOutputValueProvider
from gems.simulation.optimization import OptimizationProblem
from gems.simulation.output_values_base import BaseOutputValue
from gems.study.data import TimeScenarioIndex


@dataclass
class OutputVariable(BaseOutputValue):
    """
    Contains a single solver variable's values and status.
    All shared logic is now in BaseOutputValue.
    """

    _basis_status: Dict[TimeScenarioIndex, str] = field(
        init=False, default_factory=dict
    )

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, OutputVariable):
            return NotImplemented
        # Check base equality first (name, size, value)
        if not super().__eq__(other):
            return False
        # Then check the unique field
        return (self.ignore or other.ignore) or (
            self._basis_status == other._basis_status
        )

    def _set(
        self,
        timestep: Optional[int],
        scenario: Optional[int],
        value: float,
        status: Optional[str] = None,
        is_mip: bool = True,
    ) -> None:
        timestep = 0 if timestep is None else timestep
        scenario = 0 if scenario is None else scenario
        key = TimeScenarioIndex(timestep, scenario)
        if key not in self._value:
            size_s = max(self._size[0], scenario + 1)
            size_t = max(self._size[1], timestep + 1)
            self._size = (size_s, size_t)

        self._value[key] = value
        if not is_mip and status is not None:
            self._basis_status[key] = status


@dataclass
class OutputComponent:
    _id: str
    _variables: Dict[str, OutputVariable] = field(init=False, default_factory=dict)
    _extra_outputs: Dict[str, ExtraOutput] = field(init=False, default_factory=dict)
    model: Optional[Any] = field(default=None, init=False)
    ignore: bool = field(default=False, init=False)

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, OutputComponent):
            return NotImplemented
        return self.is_close(other, rel_tol=0.0, abs_tol=0.0)

    def is_close(
        self,
        other: "OutputComponent",
        *,
        rel_tol: float = 1.0e-9,
        abs_tol: float = 0.0,
    ) -> bool:
        return (self.ignore or other.ignore) or (
            self._id == other._id
            and _are_mappings_close(self._variables, other._variables, rel_tol, abs_tol)
            and _are_mappings_close(
                self._extra_outputs, other._extra_outputs, rel_tol, abs_tol
            )
        )

    def __str__(self) -> str:
        string = f"{self._id} : {'(ignored)' if self.ignore else ''}\n"
        for var in self._variables.values():
            string += f"  {str(var)}\n"
        if self._extra_outputs:
            string += "  [Extra Outputs]\n"
            for out in self._extra_outputs.values():
                string += f"    {out._name}: {out._value}\n"
        return string

    def var(self, variable_name: str) -> OutputVariable:
        if variable_name not in self._variables:
            self._variables[variable_name] = OutputVariable(variable_name)
        return self._variables[variable_name]

    def extra_output(self, output_name: str) -> ExtraOutput:
        if output_name not in self._extra_outputs:
            self._extra_outputs[output_name] = ExtraOutput(output_name)
        return self._extra_outputs[output_name]

    def evaluate_extra_outputs(self, problem: OptimizationProblem) -> None:
        """Evaluate all model-defined extra outputs and populate self._extra_outputs."""
        if problem is None:
            raise ValueError("Expected a valid OptimizationProblem, got None.")

        if self.model is None or self.model.extra_outputs is None:
            return

        self._extra_outputs = {}

        for out_id, expr_node in self.model.extra_outputs.items():
            if out_id not in self._extra_outputs:
                self._extra_outputs[out_id] = ExtraOutput(out_id)

            self._evaluate_single_extra_output(
                self._extra_outputs[out_id], problem, expr_node
            )

    def _evaluate_single_extra_output(
        self,
        extra_output: ExtraOutput,
        problem: OptimizationProblem,
        expr_node: Any,
    ) -> None:
        """
        Evaluate a single ExtraOutput for all time/scenario indices
        from the component's variables.
        """
        all_indices: Set[TimeScenarioIndex] = set()
        for var in self._variables.values():
            all_indices.update(var._value.keys())
        if not all_indices:
            all_indices = {TimeScenarioIndex(0, 0)}

        sorted_indices = sorted(all_indices, key=lambda k: (k.time, k.scenario))

        for idx in sorted_indices:
            try:
                expanded_expr = problem.context.expand_operators(expr_node)
                provider = ExtraOutputValueProvider(self, problem, idx)
                val = float(evaluate(expanded_expr, provider))
            except EvaluationError as e:
                print(
                    f"[ERROR] Eval failed for '{extra_output._name}' in {self._id} "
                    f"at t={idx.time}, s={idx.scenario}: {e}"
                )
                val = float("nan")
            except Exception as e:
                print(
                    f"[ERROR] Unexpected error for '{extra_output._name}' in {self._id}: {e}"
                )
                val = float("nan")

            extra_output._set(idx.time, idx.scenario, val)


@dataclass
class OutputValues:
    """
    Contains variables and extra outputs after solver work completion.
    """

    problem: Optional[OptimizationProblem] = field(default=None)
    _components: Dict[str, OutputComponent] = field(init=False, default_factory=dict)

    def __post_init__(self) -> None:
        self._build_components()
        self._fill_components()

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, OutputValues):
            return NotImplemented
        return _are_mappings_close(self._components, other._components, 0.0, 0.0)

    def is_close(
        self, other: "OutputValues", *, rel_tol: float = 1.0e-9, abs_tol: float = 0.0
    ) -> bool:
        return _are_mappings_close(
            self._components, other._components, rel_tol, abs_tol
        )

    def __str__(self) -> str:
        return "\n" + "".join(f"{comp}\n" for comp in self._components.values())

    def _build_components(self) -> None:
        """
        Initializes component objects and links them to their models.
        It only creates the structure, no values are set.
        """
        if self.problem is None:
            return

        # Ensure a Component object exists for every component in the network
        for cmp in self.problem.context.network.all_components:
            comp = self.component(cmp.id)
            comp.model = cmp.model

    def _fill_components(self) -> None:
        """
        Fills all output values (Variables from solver, ExtraOutputs from evaluation).
        """
        # 1. Populate Variables
        self._evaluate_variables()

        # 2. Evaluate Extra Outputs, which depend on the variables being set
        self._evaluate_extra_outputs()

    def component(self, component_id: str) -> OutputComponent:
        if component_id not in self._components:
            self._components[component_id] = OutputComponent(component_id)
        return self._components[component_id]

    def _evaluate_variables(self) -> None:
        """
        Populates the OutputVariable values from the solver results. # Docstring updated
        """
        if self.problem is None:
            return

        is_mip = self.problem.solver.IsMip()

        for key, value in self.problem.context.get_all_component_variables().items():
            status = None if is_mip else value.basis_status()
            self.component(key.component_id).var(str(key.variable_name))._set(
                key.block_timestep,
                key.scenario,
                value.solution_value(),
                status=status,
                is_mip=is_mip,
            )

    def _evaluate_extra_outputs(self) -> None:
        """Evaluate extra outputs for all components."""
        if self.problem is None:
            return
        for comp in self._components.values():
            comp.evaluate_extra_outputs(self.problem)


Comparable = TypeVar("Comparable", OutputComponent, OutputVariable, ExtraOutput)


def _are_mappings_close(
    lhs: Mapping[str, Comparable],
    rhs: Mapping[str, Comparable],
    rel_tol: float,
    abs_tol: float,
) -> bool:
    lhs_keys = lhs.keys()
    rhs_keys = rhs.keys()

    # Keys present only on the left
    for key in lhs_keys - rhs_keys:
        if not lhs[key].ignore:
            return False

    # Keys present only on the right
    for key in rhs_keys - lhs_keys:
        if not rhs[key].ignore:
            return False

    # Keys in common
    for key in lhs_keys & rhs_keys:
        left_item = lhs[key]
        right_item = rhs[key]
        if not left_item.ignore and not left_item.is_close(
            right_item, rel_tol=rel_tol, abs_tol=abs_tol
        ):
            return False

    return True


@dataclass(frozen=True)
class BendersSolution:
    data: Dict[str, Any]

    def __eq__(self, other: object) -> bool:
        if not isinstance(other, BendersSolution):
            return NotImplemented
        return (
            self.overall_cost == other.overall_cost
            and self.candidates == other.candidates
        )

    def is_close(
        self,
        other: "BendersSolution",
        *,
        rel_tol: float = 1.0e-9,
        abs_tol: float = 0.0,
    ) -> bool:
        return (
            math.isclose(
                self.overall_cost, other.overall_cost, abs_tol=abs_tol, rel_tol=rel_tol
            )
            and self.candidates.keys() == other.candidates.keys()
            and all(
                math.isclose(
                    self.candidates[key],
                    other.candidates[key],
                    rel_tol=rel_tol,
                    abs_tol=abs_tol,
                )
                for key in self.candidates
            )
        )

    def __str__(self) -> str:
        lpad = 30
        rpad = 12

        string = "Benders' solution:\n"
        string += f"{'Overall cost':<{lpad}} : {self.overall_cost:>{rpad}}\n"
        string += f"{'Investment cost':<{lpad}} : {self.investment_cost:>{rpad}}\n"
        string += f"{'Operational cost':<{lpad}} : {self.operational_cost:>{rpad}}\n"
        string += "-" * (lpad + rpad + 3) + "\n"
        for candidate, investment in self.candidates.items():
            string += f"{candidate:<{lpad}} : {investment:>{rpad}}\n"

        return string

    @property
    def investment_cost(self) -> float:
        return self.data["solution"]["investment_cost"]

    @property
    def operational_cost(self) -> float:
        return self.data["solution"]["operational_cost"]

    @property
    def overall_cost(self) -> float:
        return self.data["solution"]["overall_cost"]

    @property
    def candidates(self) -> Dict[str, float]:
        return self.data["solution"]["values"]

    @property
    def status(self) -> str:
        return self.data["solution"]["problem_status"]

    @property
    def absolute_gap(self) -> float:
        return self.data["solution"]["optimality_gap"]

    @property
    def relative_gap(self) -> float:
        return self.data["solution"]["relative_gap"]

    @property
    def stopping_criterion(self) -> str:
        return self.data["solution"]["stopping_criterion"]


@dataclass(frozen=True, eq=False)
class BendersMergedSolution(BendersSolution):
    @property
    def lower_bound(self) -> float:
        return self.data["solution"]["lb"]

    @property
    def upper_bound(self) -> float:
        return self.data["solution"]["ub"]


@dataclass(frozen=True, eq=False)
class BendersDecomposedSolution(BendersSolution):
    @property
    def nb_iterations(self) -> int:
        return self.data["solution"]["iteration"]

    @property
    def duration(self) -> float:
        return self.data["run_duration"]



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from dataclasses import dataclass, replace
from typing import Any

from gems.expression.indexing_structure import IndexingStructure
from gems.model.common import ValueType


@dataclass(frozen=True)
class Parameter:
    """
    A parameter of the model: a parameter is mainly defined by a name and expected type.
    When the model is instantiated as a component, a value must be provided for
    parameters, either as constant values or timeseries-based values.
    """

    name: str
    type: ValueType
    structure: IndexingStructure

    def replicate(self, /, **changes: Any) -> "Parameter":
        return replace(self, **changes)


def int_parameter(
    name: str,
    structure: IndexingStructure = IndexingStructure(True, True),
) -> Parameter:
    return Parameter(name, ValueType.INTEGER, structure)


def float_parameter(
    name: str,
    structure: IndexingStructure = IndexingStructure(True, True),
) -> Parameter:
    return Parameter(name, ValueType.CONTINUOUS, structure)



# Copyright (c) 2025, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.


import os
import subprocess
import time
from pathlib import Path
from typing import Tuple

import numpy as np
import pandas as pd


class AntaresHybridRunner:
    def __init__(
        self, exec_dir: Path, study_dir: Path, solver: None | str = None
    ) -> None:
        self.exec_dir, self.study_dir, self.solver = exec_dir, study_dir, solver
        self.exec_path = self.exec_dir / Path("antares-solver")
        self.OUTPUT_FILE_DIR_NAME = "output"
        self.SIMULATION_TABLE_1_FILE = "simulation_table--optim-nb-1.csv"
        self.SIMULATION_TABLE_2_FILE = "simulation_table--optim-nb-2.csv"

    def run(self) -> None:
        # Build the command to run antares-solver.exe
        command = [str(self.exec_path.absolute()), str(self.study_dir.absolute())]
        if self.solver:
            command.append("--linear-solver")
            command.append(self.solver)
        # Run the command and wait for it to finish
        try:
            start = time.time()
            subprocess.run(
                command,
                check=True,  # raise CalledProcessError on failure
                capture_output=True,  # capture stdout/stderr if needed
                text=True,  # decode bytes to string
            )
            # Log or inspect result.stdout / result.stderr here if desired
            end = time.time()
            self.total_execution_time = end - start
        except subprocess.CalledProcessError as e:
            raise RuntimeError(
                f"Antares execution failed with code {e.returncode}: {e.stderr}"
            )
        # Read the objective value from the solution file
        path = self.study_dir / Path(self.OUTPUT_FILE_DIR_NAME)
        folders = [
            name for name in os.listdir(path) if os.path.isdir(os.path.join(path, name))
        ]
        self.simulation_id = sorted(folders)[-1]
        simulation_table_1_path, simulation_table_2_path = path / Path(
            self.simulation_id + "/" + self.SIMULATION_TABLE_1_FILE
        ), path / Path(self.simulation_id + "/" + self.SIMULATION_TABLE_2_FILE)
        self.simulation_table_1, self.simulation_table_2 = pd.read_csv(
            simulation_table_1_path
        ), pd.read_csv(simulation_table_2_path)


class AntaresHybridStudyBenchmarker:
    def __init__(
        self,
        exec_path: Path,
        study_path_1: Path,
        study_path_2: Path,
        solver: None | str = None,
    ) -> None:
        self.exec_path, self.study_path_1, self.study_path_2, self.solver = (
            exec_path,
            study_path_1,
            study_path_2,
            solver,
        )
        self.weekly_objectives_1: list = []
        self.weekly_objectives_2: list = []
        self.simulation_tables_1: list = []
        self.simulation_tables_2: list = []

    def run(self) -> None:
        for study in [self.study_path_1, self.study_path_2]:
            r = AntaresHybridRunner(self.exec_path, study, self.solver)
            r.run()
            self.weekly_objectives_1.append(
                r.simulation_table_1.query('output == "OBJECTIVE_VALUE"')[
                    "value"
                ].values
            )
            self.weekly_objectives_2.append(
                r.simulation_table_2.query('output == "OBJECTIVE_VALUE"')[
                    "value"
                ].values
            )
            self.simulation_tables_1.append(r.simulation_table_1)
            self.simulation_tables_2.append(r.simulation_table_2)

    def weekly_rel_gaps(self) -> Tuple[np.ndarray, np.ndarray]:
        return np.abs(
            (self.weekly_objectives_1[0] - self.weekly_objectives_1[1])
            / self.weekly_objectives_1[0]
        ), np.abs(
            (self.weekly_objectives_2[0] - self.weekly_objectives_2[1])
            / self.weekly_objectives_2[0]
        )

    def weekly_abs_gaps(self) -> Tuple[np.ndarray, np.ndarray]:
        return np.abs(
            self.weekly_objectives_1[0] - self.weekly_objectives_1[1]
        ), np.abs(self.weekly_objectives_2[0] - self.weekly_objectives_2[1])


class AntaresModelerRunner:
    def __init__(self, exec_dir: str, study_dir: str) -> None:
        self.exec_dir, self.study_dir = exec_dir, study_dir
        self.exec_path = self.exec_dir / Path("antares-modeler")
        self.OUTPUT_FILE_DIR_NAME = "output"

    def run(self) -> None:
        # Build the command to run antares-modeler.exe
        command = [str(self.exec_path), str(self.study_dir)]
        # Run the command and wait for it to finish
        try:
            start = time.time()
            subprocess.run(
                command,
                check=True,  # raise CalledProcessError on failure
                capture_output=True,  # capture stdout/stderr if needed
                text=True,  # decode bytes to string
            )
            # Log or inspect result.stdout / result.stderr here if desired
            end = time.time()
            self.total_execution_time = end - start
        except subprocess.CalledProcessError as e:
            raise RuntimeError(
                f"Antares execution failed with code {e.returncode}: {e.stderr}"
            )
        # Read the objective value from the solution file
        path = self.study_dir / Path(self.OUTPUT_FILE_DIR_NAME)
        folders = [name for name in os.listdir(path)]
        self.simulation_file = sorted(folders)[-1]
        assert "simulation_table" in self.simulation_file
        simulation_table_path = path / Path(self.simulation_file)
        self.simulation_table = pd.read_csv(simulation_table_path)



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from abc import ABC, abstractmethod
from typing import Generator, Optional

from gems.expression import ExpressionNode, literal
from gems.model import Constraint, Model, ProblemContext, Variable


class ModelSelectionStrategy(ABC):
    """
    Abstract class to specify the strategy of the created problem.
    Its derived classes select variables and constraints for the optimization problems:
        - InvestmentProblemStrategy: Keep investment and coupling variables and constraints only for a BendersDecomposed master
        - OperationalProblemStrategy: Keep operational and coupling variables and constraints only for a BendersDecomposed sub-problems
        - MergedProblemStrategy: Keep all variables and constraints
    """

    def get_variables(self, model: Model) -> Generator[Variable, None, None]:
        for variable in model.variables.values():
            if self._keep_from_context(variable.context):
                yield variable

    def get_constraints(self, model: Model) -> Generator[Constraint, None, None]:
        for constraint in model.get_all_constraints():
            if self._keep_from_context(constraint.context):
                yield constraint

    @abstractmethod
    def _keep_from_context(self, context: ProblemContext) -> bool:
        ...

    @abstractmethod
    def get_objectives(
        self, model: Model
    ) -> Generator[Optional[ExpressionNode], None, None]:
        ...


class MergedProblemStrategy(ModelSelectionStrategy):
    def _keep_from_context(self, context: ProblemContext) -> bool:
        return True

    def get_objectives(
        self, model: Model
    ) -> Generator[Optional[ExpressionNode], None, None]:
        yield model.objective_operational_contribution
        yield model.objective_investment_contribution


class InvestmentProblemStrategy(ModelSelectionStrategy):
    def _keep_from_context(self, context: ProblemContext) -> bool:
        return (
            context == ProblemContext.INVESTMENT or context == ProblemContext.COUPLING
        )

    def get_objectives(
        self, model: Model
    ) -> Generator[Optional[ExpressionNode], None, None]:
        yield model.objective_investment_contribution


class OperationalProblemStrategy(ModelSelectionStrategy):
    def _keep_from_context(self, context: ProblemContext) -> bool:
        return (
            context == ProblemContext.OPERATIONAL or context == ProblemContext.COUPLING
        )

    def get_objectives(
        self, model: Model
    ) -> Generator[Optional[ExpressionNode], None, None]:
        yield model.objective_operational_contribution


class RiskManagementStrategy(ABC):
    """
    Abstract functor class for risk management
    Its derived classes will implement risk measures:
        - UniformRisk   : The default case. All expressions have the same weight
        - ExpectedValue : Computes the product prob * expression
    TODO For now, it will only take into account the Expected Value
    TODO In the future could have other risk measures?
    """

    def __call__(self, expr: ExpressionNode) -> ExpressionNode:
        return self._modify_expression(expr)

    @abstractmethod
    def _modify_expression(self, expr: ExpressionNode) -> ExpressionNode:
        ...


class UniformRisk(RiskManagementStrategy):
    def _modify_expression(self, expr: ExpressionNode) -> ExpressionNode:
        return expr


class ExpectedValue(RiskManagementStrategy):
    def __init__(self, prob: float) -> None:
        self._prob = prob

    def _modify_expression(self, expr: ExpressionNode) -> ExpressionNode:
        return literal(self._prob) * expr



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

"""
The xpansion module extends the optimization module
with Benders solver related functions
"""

import pathlib
from typing import Any, Dict, List, Optional

from gems.simulation.decision_tree import DecisionTreeNode
from gems.simulation.optimization import (
    BlockBorderManagement,
    OptimizationProblem,
    build_problem,
    fusion_problems,
)
from gems.simulation.output_values import (
    BendersDecomposedSolution,
    BendersMergedSolution,
    BendersSolution,
)
from gems.simulation.runner import BendersRunner, MergeMPSRunner
from gems.simulation.strategy import (
    ExpectedValue,
    InvestmentProblemStrategy,
    OperationalProblemStrategy,
)
from gems.simulation.time_block import TimeBlock
from gems.study.data import DataBase
from gems.utils import read_json, serialize, serialize_json


class BendersDecomposedProblem:
    """
    A simpler interface for the Benders Decomposed problem
    """

    master: OptimizationProblem
    subproblems: List[OptimizationProblem]

    emplacement: pathlib.Path
    output_path: pathlib.Path
    structure_filename: str

    solution: Optional[BendersSolution]
    is_merged: bool

    def __init__(
        self,
        master: OptimizationProblem,
        subproblems: List[OptimizationProblem],
        emplacement: str = "outputs/lp",
        output_path: str = "expansion",
        struct_filename: str = "structure.txt",
    ) -> None:
        self.master = master
        self.subproblems = subproblems

        self.emplacement = pathlib.Path(emplacement)
        self.output_path = pathlib.Path(output_path)
        self.structure_filename = struct_filename

        self.solution = None
        self.is_merged = False

    def export_structure(self) -> str:
        """
        Write the structure.txt file
        """

        if not self.subproblems:
            raise RuntimeError("Subproblem list must have at least one sub problem")

        # A mapping similar to the Xpansion mapping for keeping track of variable indexes
        # in Master and Sub-problem files
        problem_to_candidates: Dict[str, Dict[str, int]] = {}
        candidates = set()

        problem_to_candidates["master"] = {}
        for solver_var_info in self.master.context._solver_variables.values():
            problem_to_candidates["master"][
                solver_var_info.name
            ] = solver_var_info.column_id
            candidates.add(solver_var_info.name)

        for problem in self.subproblems:
            problem_to_candidates[problem.name] = {}

            for solver_var_info in problem.context._solver_variables.values():
                if solver_var_info.name in candidates:
                    # If candidate was identified in master
                    problem_to_candidates[problem.name][
                        solver_var_info.name
                    ] = solver_var_info.column_id

        structure_str = ""
        for problem_name, candidate_to_index in problem_to_candidates.items():
            for candidate, index in candidate_to_index.items():
                structure_str += f"{problem_name:>50}{candidate:>50}{index:>10}\n"

        return structure_str

    def export_options(
        self, *, solver_name: str = "XPRESS", log_level: int = 0
    ) -> Dict[str, Any]:
        # Default values
        options_values = {
            "MAX_ITERATIONS": -1,
            "ABSOLUTE_GAP": 1,
            "RELATIVE_GAP": 1e-6,
            "RELAXED_GAP": 1e-5,
            "AGGREGATION": False,
            "OUTPUTROOT": ".",
            "TRACE": True,
            "SLAVE_WEIGHT": "CONSTANT",
            "SLAVE_WEIGHT_VALUE": 1,
            "MASTER_NAME": f"{self.master.name}",
            "STRUCTURE_FILE": f"{self.structure_filename}",
            "INPUTROOT": ".",
            "CSV_NAME": "benders_output_trace",
            "BOUND_ALPHA": True,
            "SEPARATION_PARAM": 0.5,
            "BATCH_SIZE": 0,
            "JSON_FILE": f"{self.output_path}/out.json",
            "LAST_ITERATION_JSON_FILE": f"{self.output_path}/last_iteration.json",
            "MASTER_FORMULATION": "integer",
            "SOLVER_NAME": solver_name,
            "TIME_LIMIT": 1_000_000_000_000,
            "LOG_LEVEL": log_level,
            "LAST_MASTER_MPS": "master_last_iteration",
            "LAST_MASTER_BASIS": "master_last_basis.bss",
        }
        return options_values

    def initialise(
        self,
        *,
        solver_name: str = "XPRESS",
        log_level: int = 0,
        is_debug: bool = False,
    ) -> None:
        serialize(
            f"{self.master.name}.mps", self.master.export_as_mps(), self.emplacement
        )
        for subproblem in self.subproblems:
            serialize(
                f"{subproblem.name}.mps", subproblem.export_as_mps(), self.emplacement
            )
        serialize(
            f"{self.structure_filename}", self.export_structure(), self.emplacement
        )
        serialize_json(
            "options.json",
            self.export_options(solver_name=solver_name, log_level=log_level),
            self.emplacement,
        )

        if is_debug:
            serialize(
                f"{self.master.name}.lp", self.master.export_as_lp(), self.emplacement
            )
            for subproblem in self.subproblems:
                serialize(
                    f"{subproblem.name}.lp", subproblem.export_as_lp(), self.emplacement
                )

    def read_solution(self) -> None:
        try:
            data = read_json("out.json", self.emplacement / self.output_path)

        except FileNotFoundError:
            # TODO For now, it will return as if nothing is wrong
            # modify it with runner's run
            print("Return without reading it for now")
            return

        if self.is_merged:
            self.solution = BendersMergedSolution(data)
        else:
            self.solution = BendersDecomposedSolution(data)

    def run(
        self,
        *,
        solver_name: str = "XPRESS",
        log_level: int = 0,
        should_merge: bool = False,
        show_debug: bool = False,
    ) -> bool:
        self.initialise(
            solver_name=solver_name, log_level=log_level, is_debug=show_debug
        )

        if not should_merge:
            return_code = BendersRunner(self.emplacement).run()
        else:
            self.is_merged = True
            return_code = MergeMPSRunner(self.emplacement).run()

        if return_code == 0:
            self.read_solution()
            return True
        else:
            return False


def build_benders_decomposed_problem(
    decision_tree_root: DecisionTreeNode,
    database: DataBase,
    *,
    border_management: BlockBorderManagement = BlockBorderManagement.CYCLE,
    solver_id: str = "GLOP",
    struct_filename: str = "structure.txt",
) -> BendersDecomposedProblem:
    """
    Entry point to build the xpansion pathway problem.

    For each node of the tree, it builds a Master and a Subproblem.
    Then it defines a coupled problem that merges all masters along with
    its pathway constraints into one 'tree master' problem.

    Returns a Benders Decomposed problem
    """

    if not decision_tree_root.is_leaves_prob_sum_one():
        raise ValueError("Decision tree leaves' probability must sum one!")

    null_time_block = TimeBlock(
        0, [0]
    )  # Not necessary for master, but list must be non-empty
    null_scenario = 0  # Not necessary for master

    decision_tree_root._add_coupling_constraints()

    coupler = build_problem(
        decision_tree_root.coupling_network,
        database,
        null_time_block,
        null_scenario,
        problem_name="coupler",
        solver_id=solver_id,
        build_strategy=InvestmentProblemStrategy(),
        risk_strategy=ExpectedValue(0.0),
        use_full_var_name=False,
    )

    masters = []  # Benders Decomposed Master Problem
    subproblems = []  # Benders Decomposed Sub-problems

    for tree_node in decision_tree_root.traverse():
        suffix_tree = f"_{tree_node.id}" if decision_tree_root.size > 1 else ""

        masters.append(
            build_problem(
                tree_node.network,
                database,
                null_time_block,
                null_scenario,
                problem_name=f"master{suffix_tree}",
                solver_id=solver_id,
                build_strategy=InvestmentProblemStrategy(),
                decision_tree_node=tree_node.id,
                risk_strategy=ExpectedValue(tree_node.prob),
            )
        )

        for block in tree_node.config.blocks:
            suffix_block = f"_b{block.id}" if len(tree_node.config.blocks) > 1 else ""

            subproblems.append(
                build_problem(
                    tree_node.network,
                    database,
                    block,
                    tree_node.config.scenarios,
                    problem_name=f"subproblem{suffix_tree}{suffix_block}",
                    solver_id=solver_id,
                    build_strategy=OperationalProblemStrategy(),
                    decision_tree_node=tree_node.id,
                    risk_strategy=ExpectedValue(tree_node.prob),
                )
            )

    master = fusion_problems(masters, coupler)

    return BendersDecomposedProblem(
        master, subproblems, struct_filename=struct_filename
    )



# Copyright (c) 2025, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from dataclasses import dataclass, field
from typing import Optional, TextIO, Union

import pandas as pd
from pydantic import BaseModel, Field
from yaml import safe_load


@dataclass
class VirtualObjectsRepository:
    areas: list[str] = field(default_factory=list)
    links: list[str] = field(default_factory=list)
    thermals: list[str] = field(default_factory=list)

    def resolve_template(
        self, template_pattern: str, value: str
    ) -> "VirtualObjectsRepository":
        thermals = []
        for thermal in self.thermals:
            thermals.append(thermal.replace(template_pattern, value))
        return VirtualObjectsRepository(
            areas=self.areas, links=self.links, thermals=thermals
        )

    def add(self, other: "VirtualObjectsRepository") -> None:
        self.areas += other.areas
        self.links += other.links
        self.thermals += other.thermals


def _to_kebab(snake: str) -> str:
    return snake.replace("_", "-")


class ModifiedBaseModel(BaseModel):
    class Config:
        alias_generator = _to_kebab
        extra = "forbid"
        populate_by_name = True


class Operation(ModifiedBaseModel):
    type: Optional[str] = None
    multiply_by: Optional[Union[float, str]] = None
    divide_by: Optional[Union[float, str]] = None

    def execute(
        self,
        initial_value: Union[pd.DataFrame, pd.Series, float],
        preprocessed_values: Optional[Union[dict[str, float], float]] = None,
    ) -> Union[float, pd.Series, pd.DataFrame]:
        def resolve(value: Union[str, float]) -> Union[float, pd.Series]:
            if isinstance(value, str):
                if (
                    not isinstance(preprocessed_values, dict)
                    or value not in preprocessed_values
                ):
                    raise ValueError(
                        f"Missing value for key '{value}' in preprocessed_values"
                    )
                return preprocessed_values[value]
            return value

        if self.type == "max":
            return float(max(initial_value))  # type: ignore

        if self.multiply_by is not None:
            return initial_value * resolve(self.multiply_by)

        if self.divide_by is not None:
            return initial_value / resolve(self.divide_by)

        raise ValueError(
            "Operation must have at least one of 'multiply_by', 'divide_by', or 'type'"
        )


class ObjectProperties(ModifiedBaseModel):
    type: str
    area: Optional[str] = None
    link: Optional[str] = None
    cluster: Optional[str] = None
    binding_constraint_id: Optional[str] = None
    field: Optional[str] = None

    def resolve_template(self, template_pattern: str, value: str) -> "ObjectProperties":
        area = (
            self.area.replace(template_pattern, value)
            if self.area is not None
            else None
        )
        link = (
            self.link.replace(template_pattern, value)
            if self.link is not None
            else None
        )
        cluster = (
            self.cluster.replace(template_pattern, value)
            if self.cluster is not None
            else None
        )
        binding_constraint_id = (
            self.binding_constraint_id.replace(template_pattern, value)
            if self.binding_constraint_id is not None
            else None
        )
        field = (
            self.field.replace(template_pattern, value)
            if self.field is not None
            else None
        )
        return ObjectProperties(
            type=self.type,
            area=area,
            link=link,
            cluster=cluster,
            binding_constraint_id=binding_constraint_id,
            field=field,
        )


class ConversionValue(ModifiedBaseModel):
    object_properties: Optional[ObjectProperties] = None
    column: Optional[int] = None
    operation: Optional[Operation] = None
    constant: Optional[float] = None

    def resolve_template(self, template_pattern: str, value: str) -> "ConversionValue":
        object_properties = (
            self.object_properties.resolve_template(template_pattern, value)
            if self.object_properties is not None
            else None
        )
        return ConversionValue(
            object_properties=object_properties,
            column=self.column,
            operation=self.operation,
            constant=self.constant,
        )


class ParameterConversionConfig(ModifiedBaseModel):
    id: str
    time_dependent: bool = False
    scenario_dependent: bool = False
    value: ConversionValue

    def resolve_template(
        self, template_pattern: str, value: str
    ) -> "ParameterConversionConfig":
        return ParameterConversionConfig(
            id=self.id,
            time_dependent=self.time_dependent,
            scenario_dependent=self.scenario_dependent,
            value=self.value.resolve_template(template_pattern, value),
        )


class ComponentConversionConfig(ModifiedBaseModel):
    id: str
    parameters: list[ParameterConversionConfig]

    def resolve_template(
        self, template_pattern: str, value: str
    ) -> "ComponentConversionConfig":
        id = self.id.replace(template_pattern, value)
        parameters = []
        for param in self.parameters:
            parameters.append(param.resolve_template(template_pattern, value))
        return ComponentConversionConfig(id=id, parameters=parameters)


class ReferencedLegacyObjects(ModifiedBaseModel):
    id: str
    object_properties: ObjectProperties

    def resolve_template(
        self, template_pattern: str, value: str
    ) -> "ReferencedLegacyObjects":
        return ReferencedLegacyObjects(
            id=self.id,
            object_properties=self.object_properties.resolve_template(
                template_pattern, value
            ),
        )


class AreaConnectionConversionConfig(ModifiedBaseModel):
    component: str
    port: str
    area: str

    def resolve_template(
        self, template_pattern: str, value: str
    ) -> "AreaConnectionConversionConfig":
        return AreaConnectionConversionConfig(
            component=self.component.replace(template_pattern, value),
            port=self.port,
            area=self.area.replace(template_pattern, value),
        )


class PortConnectionConversionConfig(ModifiedBaseModel):
    component1: str
    port1: str
    component2: str
    port2: str

    def resolve_template(
        self, template_pattern: str, value: str
    ) -> "PortConnectionConversionConfig":
        return PortConnectionConversionConfig(
            component1=self.component1.replace(template_pattern, value),
            port1=self.port1,
            component2=self.component2.replace(template_pattern, value),
            port2=self.port2,
        )


class TemplateParameter(ModifiedBaseModel):
    name: str
    description: Optional[str] = None
    cluster_type: Optional[str] = None
    exclude: Optional[list[ReferencedLegacyObjects]] = None


class ConversionTemplate(ModifiedBaseModel):
    name: str
    model: str
    generator_version_compatibility: Optional[str] = None
    template_parameters: list[TemplateParameter] = Field(default_factory=list)
    component: ComponentConversionConfig
    connections: list[PortConnectionConversionConfig] = Field(default_factory=list)
    area_connections: list[AreaConnectionConversionConfig] = Field(default_factory=list)
    legacy_objects_to_delete: list[ReferencedLegacyObjects] = Field(
        default_factory=list
    )
    scenario_group: Optional[str] = None

    def resolve_template(
        self, template_pattern: str, value: str
    ) -> "ConversionTemplate":
        component = self.component.resolve_template(template_pattern, value)
        connections = []
        area_connections = []
        legacy_objects_to_delete = []
        for connection in self.connections:
            connections.append(connection.resolve_template(template_pattern, value))
        for area_connection in self.area_connections:
            area_connections.append(
                area_connection.resolve_template(template_pattern, value)
            )
        for object_to_delete in self.legacy_objects_to_delete:
            legacy_objects_to_delete.append(
                object_to_delete.resolve_template(template_pattern, value)
            )
        return ConversionTemplate(
            name=self.name,
            model=self.model,
            generator_version_compatibility=self.generator_version_compatibility,
            template_parameters=self.template_parameters,
            component=component,
            connections=connections,
            area_connections=area_connections,
            legacy_objects_to_delete=legacy_objects_to_delete,
            scenario_group=self.scenario_group,
        )

    def get_excluded_objects_ids(self) -> VirtualObjectsRepository:
        virtual_areas = []
        virtual_links = []
        virtual_thermals = []
        for param in self.template_parameters:
            if param.exclude is not None:
                for excluded_object in param.exclude:
                    exclude_prop = excluded_object.object_properties
                    if exclude_prop.type == "area" and exclude_prop.area is not None:
                        virtual_areas.append(exclude_prop.area)
                    elif exclude_prop.type == "link" and exclude_prop.link is not None:
                        virtual_links.append(exclude_prop.link)
                    elif exclude_prop.type == "thermal":
                        if (
                            exclude_prop.area is not None
                            and exclude_prop.cluster is not None
                        ):
                            virtual_thermals.append(exclude_prop.cluster)
        return VirtualObjectsRepository(virtual_areas, virtual_links, virtual_thermals)


def parse_conversion_template(input_template: TextIO) -> ConversionTemplate:
    tree = safe_load(input_template)
    return ConversionTemplate.model_validate(tree["template"])



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from dataclasses import dataclass, field, replace
from typing import Any

from gems.expression.degree import is_constant
from gems.expression.equality import expressions_equal, expressions_equal_if_present
from gems.expression.expression import (
    Comparator,
    ComparisonNode,
    ExpressionNode,
    is_non_negative,
    is_unbounded,
    literal,
)
from gems.expression.print import print_expr
from gems.model.common import ProblemContext


@dataclass
class Constraint:
    """
    A constraint linking variables and parameters of a model together.

    No variable is expected on the right hand side of the constraint.
    """

    name: str
    expression: ExpressionNode
    lower_bound: ExpressionNode = field(default=literal(-float("inf")))
    upper_bound: ExpressionNode = field(default=literal(float("inf")))
    context: ProblemContext = field(default=ProblemContext.OPERATIONAL)

    def __post_init__(
        self,
    ) -> None:
        if isinstance(self.expression, ComparisonNode):
            if not is_unbounded(self.lower_bound) or not is_unbounded(self.upper_bound):
                raise ValueError(
                    "Both comparison between two expressions and a bound are specfied, set either only a comparison between expressions or a single linear expression with bounds."
                )

            if self.expression.comparator == Comparator.LESS_THAN:
                # lhs - rhs <= 0
                self.upper_bound = literal(0)
                self.lower_bound = literal(-float("inf"))
            elif self.expression.comparator == Comparator.GREATER_THAN:
                # lhs - rhs >= 0
                self.lower_bound = literal(0)
                self.upper_bound = literal(float("inf"))
            else:  # lhs - rhs == 0
                self.lower_bound = literal(0)
                self.upper_bound = literal(0)

            self.expression = self.expression.left - self.expression.right

        else:
            for bound in [self.lower_bound, self.upper_bound]:
                if not is_constant(bound):
                    raise ValueError(
                        f"The bounds of a constraint should not contain variables, {print_expr(bound)} was given."
                    )

            if is_unbounded(self.lower_bound) and is_non_negative(self.lower_bound):
                raise ValueError("Lower bound should not be +Inf")

            if is_unbounded(self.upper_bound) and not is_non_negative(self.upper_bound):
                raise ValueError("Upper bound should not be -Inf")

    def replicate(self, /, **changes: Any) -> "Constraint":
        return replace(self, **changes)

    def __eq__(self, other: Any) -> bool:
        if not isinstance(other, Constraint):
            return False
        return (
            self.name == other.name
            and expressions_equal(self.expression, other.expression)
            and expressions_equal_if_present(self.lower_bound, other.lower_bound)
            and expressions_equal_if_present(self.upper_bound, other.upper_bound)
        )



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.
from dataclasses import dataclass
from typing import Dict, Iterable

from gems.model import Model, PortType


@dataclass(frozen=True)
class Library:
    id: str
    port_types: Dict[str, PortType]
    models: Dict[str, Model]


def library(
    id: str,
    port_types: Iterable[PortType],
    models: Iterable[Model],
) -> Library:
    return Library(
        id=id,
        port_types=dict((p.id, p) for p in port_types),
        models=dict((m.id, m) for m in models),
    )



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from dataclasses import dataclass
from typing import List, cast

from .expression import (
    AllTimeSumNode,
    ComparisonNode,
    ComponentParameterNode,
    ComponentVariableNode,
    ExpressionNode,
    LiteralNode,
    ParameterNode,
    PortFieldAggregatorNode,
    PortFieldNode,
    ProblemParameterNode,
    ProblemVariableNode,
    ScenarioOperatorNode,
    TimeEvalNode,
    TimeShiftNode,
    TimeSumNode,
    VariableNode,
)
from .visitor import ExpressionVisitorOperations, visit


@dataclass(frozen=True)
class CopyVisitor(ExpressionVisitorOperations[ExpressionNode]):
    """
    Simply copies the whole AST.
    """

    def literal(self, node: LiteralNode) -> ExpressionNode:
        return LiteralNode(node.value)

    def comparison(self, node: ComparisonNode) -> ExpressionNode:
        return ComparisonNode(
            visit(node.left, self), visit(node.right, self), node.comparator
        )

    def variable(self, node: VariableNode) -> ExpressionNode:
        return VariableNode(node.name)

    def parameter(self, node: ParameterNode) -> ExpressionNode:
        return ParameterNode(node.name)

    def comp_variable(self, node: ComponentVariableNode) -> ExpressionNode:
        return ComponentVariableNode(node.component_id, node.name)

    def comp_parameter(self, node: ComponentParameterNode) -> ExpressionNode:
        return ComponentParameterNode(node.component_id, node.name)

    def pb_variable(self, node: ProblemVariableNode) -> ExpressionNode:
        return ProblemVariableNode(
            node.component_id, node.name, node.time_index, node.scenario_index
        )

    def pb_parameter(self, node: ProblemParameterNode) -> ExpressionNode:
        return ProblemParameterNode(
            node.component_id, node.name, node.time_index, node.scenario_index
        )

    def time_shift(self, node: TimeShiftNode) -> ExpressionNode:
        return TimeShiftNode(visit(node.operand, self), visit(node.time_shift, self))

    def time_eval(self, node: TimeEvalNode) -> ExpressionNode:
        return TimeEvalNode(visit(node.operand, self), visit(node.eval_time, self))

    def time_sum(self, node: TimeSumNode) -> ExpressionNode:
        return TimeSumNode(
            visit(node.operand, self),
            visit(node.from_time, self),
            visit(node.to_time, self),
        )

    def all_time_sum(self, node: AllTimeSumNode) -> ExpressionNode:
        return AllTimeSumNode(visit(node.operand, self))

    def scenario_operator(self, node: ScenarioOperatorNode) -> ExpressionNode:
        return ScenarioOperatorNode(visit(node.operand, self), node.name)

    def port_field(self, node: PortFieldNode) -> ExpressionNode:
        return PortFieldNode(node.port_name, node.field_name)

    def port_field_aggregator(self, node: PortFieldAggregatorNode) -> ExpressionNode:
        return PortFieldAggregatorNode(visit(node.operand, self), node.aggregator)


def copy_expression(expression: ExpressionNode) -> ExpressionNode:
    return visit(expression, CopyVisitor())



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.
from dataclasses import dataclass
from typing import Set

from antlr4 import CommonTokenStream, InputStream
from antlr4.error.ErrorStrategy import BailErrorStrategy

from gems.expression import ExpressionNode, literal, param, var
from gems.expression.equality import expressions_equal
from gems.expression.expression import (
    Comparator,
    ComparisonNode,
    PortFieldAggregatorNode,
    PortFieldNode,
)
from gems.expression.parsing.antlr.ExprLexer import ExprLexer
from gems.expression.parsing.antlr.ExprParser import ExprParser
from gems.expression.parsing.antlr.ExprVisitor import ExprVisitor


@dataclass(frozen=True)
class ModelIdentifiers:
    """
    Allows to distinguish between parameters and variables.
    """

    variables: Set[str]
    parameters: Set[str]

    def is_variable(self, identifier: str) -> bool:
        return identifier in self.variables

    def is_parameter(self, identifier: str) -> bool:
        return identifier in self.parameters


@dataclass(frozen=True)
class ExpressionNodeBuilderVisitor(ExprVisitor):
    """
    Visits a tree created by ANTLR to create our AST representation.
    """

    identifiers: ModelIdentifiers

    # Visit a parse tree produced by ExprParser#portFieldExpr.
    def visitPortFieldExpr(
        self, ctx: ExprParser.PortFieldExprContext
    ) -> ExpressionNode:
        return PortFieldNode(
            port_name=ctx.IDENTIFIER(0).getText(),  # type: ignore
            field_name=ctx.IDENTIFIER(1).getText(),  # type: ignore
        )

    def visitFullexpr(self, ctx: ExprParser.FullexprContext) -> ExpressionNode:
        return ctx.expr().accept(self)  # type: ignore

    # Visit a parse tree produced by ExprParser#number.
    def visitNumber(self, ctx: ExprParser.NumberContext) -> ExpressionNode:
        return literal(float(ctx.NUMBER().getText()))  # type: ignore

    # Visit a parse tree produced by ExprParser#identifier.
    def visitIdentifier(self, ctx: ExprParser.IdentifierContext) -> ExpressionNode:
        return self._convert_identifier(ctx.IDENTIFIER().getText())  # type: ignore

    # Visit a parse tree produced by ExprParser#division.
    def visitMuldiv(self, ctx: ExprParser.MuldivContext) -> ExpressionNode:
        left = ctx.expr(0).accept(self)  # type: ignore
        right = ctx.expr(1).accept(self)  # type: ignore
        op = ctx.op.text  # type: ignore
        if op == "*":
            return left * right
        elif op == "/":
            return left / right
        raise ValueError(f"Invalid operator {op}")

    # Visit a parse tree produced by ExprParser#subtraction.
    def visitAddsub(self, ctx: ExprParser.AddsubContext) -> ExpressionNode:
        left = ctx.expr(0).accept(self)  # type: ignore
        right = ctx.expr(1).accept(self)  # type: ignore
        op = ctx.op.text  # type: ignore
        if op == "+":
            return left + right
        elif op == "-":
            return left - right
        raise ValueError(f"Invalid operator {op}")

    # Visit a parse tree produced by ExprParser#negation.
    def visitNegation(self, ctx: ExprParser.NegationContext) -> ExpressionNode:
        return -ctx.expr().accept(self)  # type: ignore

    # Visit a parse tree produced by ExprParser#expression.
    def visitExpression(self, ctx: ExprParser.ExpressionContext) -> ExpressionNode:
        return ctx.expr().accept(self)  # type: ignore

    # Visit a parse tree produced by ExprParser#unsignedAtom.
    def visitUnsignedAtom(self, ctx: ExprParser.UnsignedAtomContext) -> ExpressionNode:
        return ctx.atom().accept(self)  # type: ignore

    def _convert_identifier(self, identifier: str) -> ExpressionNode:
        if self.identifiers.is_variable(identifier):
            return var(identifier)
        elif self.identifiers.is_parameter(identifier):
            return param(identifier)
        raise ValueError(f"{identifier} is not a valid variable or parameter name.")

    # Visit a parse tree produced by ExprParser#portField.
    def visitPortField(self, ctx: ExprParser.PortFieldContext) -> ExpressionNode:
        return ctx.portFieldExpr().accept(self)  # type: ignore

    # Visit a parse tree produced by ExprParser#comparison.
    def visitComparison(self, ctx: ExprParser.ComparisonContext) -> ExpressionNode:
        op = ctx.COMPARISON().getText()  # type: ignore
        exp1 = ctx.expr(0).accept(self)  # type: ignore
        exp2 = ctx.expr(1).accept(self)  # type: ignore
        comp = {
            "=": Comparator.EQUAL,
            "<=": Comparator.LESS_THAN,
            ">=": Comparator.GREATER_THAN,
        }[op]
        return ComparisonNode(exp1, exp2, comp)

    # Visit a parse tree produced by ExprParser#portFieldSum.
    def visitPortFieldSum(self, ctx: ExprParser.PortFieldSumContext) -> ExpressionNode:
        return PortFieldAggregatorNode(ctx.portFieldExpr().accept(self), "PortSum")  # type: ignore

    # Visit a parse tree produced by ExprParser#timeShift.
    def visitTimeIndex(self, ctx: ExprParser.TimeIndexContext) -> ExpressionNode:
        expr = self._convert_identifier(ctx.IDENTIFIER().getText())  # type: ignore
        eval_time = ctx.expr().accept(self)  # type: ignore
        return expr.eval(eval_time)

    def visitTimeShift(self, ctx: ExprParser.TimeShiftContext) -> ExpressionNode:
        shifted_expr = self._convert_identifier(ctx.IDENTIFIER().getText())  # type: ignore
        time_shift = ctx.shift().accept(self)  # type: ignore
        # specifics for x[t] ...
        if expressions_equal(time_shift, literal(0)):
            return shifted_expr
        return shifted_expr.shift(time_shift)

    def visitTimeSum(self, ctx: ExprParser.TimeSumContext) -> ExpressionNode:
        shifted_expr = ctx.expr().accept(self)  # type: ignore
        from_shift = ctx.from_.accept(self)  # type: ignore
        to_shift = ctx.to.accept(self)  # type: ignore
        return shifted_expr.time_sum(from_shift, to_shift)

    def visitAllTimeSum(self, ctx: ExprParser.AllTimeSumContext) -> ExpressionNode:
        shifted_expr = ctx.expr().accept(self)  # type: ignore
        return shifted_expr.time_sum()

    # Visit a parse tree produced by ExprParser#function.
    def visitFunction(self, ctx: ExprParser.FunctionContext) -> ExpressionNode:
        function_name: str = ctx.IDENTIFIER().getText()  # type: ignore
        operand: ExpressionNode = ctx.expr().accept(self)  # type: ignore
        fn = _FUNCTIONS.get(function_name, None)
        if fn is None:
            raise ValueError(f"Encountered invalid function name {function_name}")
        return fn(operand)

    # Visit a parse tree produced by ExprParser#shift.
    def visitShift(self, ctx: ExprParser.ShiftContext) -> ExpressionNode:
        if ctx.shift_expr() is None:  # type: ignore
            return literal(0)
        shift = ctx.shift_expr().accept(self)  # type: ignore
        return shift

    # Visit a parse tree produced by ExprParser#shiftAddsub.
    def visitShiftAddsub(self, ctx: ExprParser.ShiftAddsubContext) -> ExpressionNode:
        left = ctx.shift_expr().accept(self)  # type: ignore
        right = ctx.right_expr().accept(self)  # type: ignore
        op = ctx.op.text  # type: ignore
        if op == "+":
            return left + right
        elif op == "-":
            return left - right
        raise ValueError(f"Invalid operator {op}")

    # Visit a parse tree produced by ExprParser#shiftMuldiv.
    def visitShiftMuldiv(self, ctx: ExprParser.ShiftMuldivContext) -> ExpressionNode:
        left = ctx.shift_expr().accept(self)  # type: ignore
        right = ctx.right_expr().accept(self)  # type: ignore
        op = ctx.op.text  # type: ignore
        if op == "*":
            return left * right
        elif op == "/":
            return left / right
        raise ValueError(f"Invalid operator {op}")

    # Visit a parse tree produced by ExprParser#signedExpression.
    def visitSignedExpression(
        self, ctx: ExprParser.SignedExpressionContext
    ) -> ExpressionNode:
        if ctx.op.text == "-":  # type: ignore
            return -ctx.expr().accept(self)  # type: ignore
        else:
            return ctx.expr().accept(self)  # type: ignore

    # Visit a parse tree produced by ExprParser#signedAtom.
    def visitSignedAtom(self, ctx: ExprParser.SignedAtomContext) -> ExpressionNode:
        if ctx.op.text == "-":  # type: ignore
            return -ctx.atom().accept(self)  # type: ignore
        else:
            return ctx.atom().accept(self)  # type: ignore

    # Visit a parse tree produced by ExprParser#rightExpression.
    def visitRightExpression(
        self, ctx: ExprParser.RightExpressionContext
    ) -> ExpressionNode:
        return ctx.expr().accept(self)  # type: ignore

    # Visit a parse tree produced by ExprParser#rightMuldiv.
    def visitRightMuldiv(self, ctx: ExprParser.RightMuldivContext) -> ExpressionNode:
        left = ctx.right_expr(0).accept(self)  # type: ignore
        right = ctx.right_expr(1).accept(self)  # type: ignore
        op = ctx.op.text  # type: ignore
        if op == "*":
            return left * right
        elif op == "/":
            return left / right
        raise ValueError(f"Invalid operator {op}")

    # Visit a parse tree produced by ExprParser#rightAtom.
    def visitRightAtom(self, ctx: ExprParser.RightAtomContext) -> ExpressionNode:
        return ctx.atom().accept(self)  # type: ignore


_FUNCTIONS = {
    "expec": ExpressionNode.expec,
}


class AntaresParseException(Exception):
    pass


def parse_expression(expression: str, identifiers: ModelIdentifiers) -> ExpressionNode:
    """
    Parses a string expression to create the corresponding AST representation.
    """
    try:
        input = InputStream(expression)
        lexer = ExprLexer(input)
        stream = CommonTokenStream(lexer)
        parser = ExprParser(stream)
        parser._errHandler = BailErrorStrategy()

        return ExpressionNodeBuilderVisitor(identifiers).visit(parser.fullexpr())  # type: ignore

    except ValueError as e:
        raise AntaresParseException(f"An error occurred during parsing: {e}") from e
    except Exception as e:
        raise AntaresParseException(
            f"An error occurred during parsing: {type(e).__name__}"
        ) from e



from dataclasses import dataclass
from typing import Any, Dict, Optional

from gems.expression.evaluate import ValueProvider
from gems.simulation.optimization import OptimizationProblem
from gems.simulation.output_values_base import BaseOutputValue
from gems.study.data import ComponentParameterIndex, TimeScenarioIndex


@dataclass
class ExtraOutput(BaseOutputValue):
    """
    Stores evaluated outputs (from ExpressionNodes), not solver variables.
    Inherits all common fields (_name, _value, _size, ignore) and methods
    (__eq__, is_close, __str__, value property) from BaseOutputValue.
    """

    def _set(
        self,
        timestep: Optional[int],
        scenario: Optional[int],
        value: float,
    ) -> None:
        timestep = 0 if timestep is None else timestep
        scenario = 0 if scenario is None else scenario
        key = TimeScenarioIndex(timestep, scenario)

        if key not in self._value:
            size_s = max(self._size[0], scenario + 1)
            size_t = max(self._size[1], timestep + 1)
            self._size = (size_s, size_t)

        self._value[key] = value


class ExtraOutputValueProvider(ValueProvider):
    # ... (content remains the same, as it only interacts with public methods/inherited fields like _value) ...
    """Provides variable and parameter values for extra output expressions."""

    def __init__(
        self,
        component: Any,
        problem: OptimizationProblem,
        idx: TimeScenarioIndex,
    ) -> None:
        self.component = component
        self.problem = problem
        self.idx = idx
        self.context = self._build_context()

    def _build_context(self) -> Dict[str, float]:
        ctx: Dict[str, float] = {}

        # --- Variables ---
        if hasattr(self.component, "_variables"):
            for vname, vobj in self.component._variables.items():
                if hasattr(vobj, "_value"):
                    val = vobj._value.get(self.idx)
                    if val is not None:
                        ctx[vname] = val
                        if hasattr(self.component, "_id"):
                            ctx[f"{self.component._id}.{vname}"] = val

        # --- Parameters ---
        model = getattr(self.component, "model", None)
        if model is not None and hasattr(model, "parameters"):
            for pname in model.parameters:
                try:
                    val = self.problem.context.database.get_value(
                        ComponentParameterIndex(self.component._id, pname),
                        self.idx.time,
                        self.idx.scenario,
                    )
                    ctx[pname] = val
                    if hasattr(self.component, "_id"):
                        ctx[f"{self.component._id}.{pname}"] = val
                except KeyError:
                    continue

        return ctx

    # ValueProvider interface
    def get_variable_value(self, name: str) -> float:
        return self.context[name]

    def get_parameter_value(self, name: str) -> float:
        return self.context[name]

    def get_component_variable_value(self, component_id: str, name: str) -> float:
        return self.context[name]

    def get_component_parameter_value(self, component_id: str, name: str) -> float:
        return self.context[name]



import dataclasses
from dataclasses import dataclass
from typing import Callable, TypeVar, Union

from gems.expression import CopyVisitor, ExpressionNode, sum_expressions, visit
from gems.expression.expression import (
    AllTimeSumNode,
    ComponentParameterNode,
    ComponentVariableNode,
    CurrentScenarioIndex,
    NoScenarioIndex,
    NoTimeIndex,
    OneScenarioIndex,
    ProblemParameterNode,
    ProblemVariableNode,
    ScenarioOperatorNode,
    TimeEvalNode,
    TimeShift,
    TimeShiftNode,
    TimeStep,
    TimeSumNode,
    problem_param,
    problem_var,
)
from gems.expression.indexing import IndexingStructureProvider

ExpressionEvaluator = Callable[[ExpressionNode], int]


@dataclass(frozen=True)
class ProblemDimensions:
    """
    Dimensions for the simulation window
    """

    timesteps_count: int
    scenarios_count: int


@dataclass(frozen=True)
class ProblemIndex:
    """
    Index of an object in the simulation window.
    """

    timestep: int
    scenario: int


@dataclass(frozen=True)
class OperatorsExpansion(CopyVisitor):
    """
    Replaces aggregators (time sum, expectations ...) by their
    arithmetic expansion.

    This will allow to easily translate it to a plain linear expression later on,
    without complex handling of operators.

    The obtained expression only contains `ProblemVariableNode` for variables
    and `ProblemParameterNode` parameters.
    """

    timesteps_count: int
    scenarios_count: int
    evaluator: ExpressionEvaluator
    structure_provider: IndexingStructureProvider

    def comp_variable(self, node: ComponentVariableNode) -> ExpressionNode:
        structure = self.structure_provider.get_component_variable_structure(
            node.component_id, node.name
        )
        time_index = TimeShift(0) if structure.time else NoTimeIndex()
        scenario_index = (
            CurrentScenarioIndex() if structure.scenario else NoScenarioIndex()
        )
        return problem_var(node.component_id, node.name, time_index, scenario_index)

    def comp_parameter(self, node: ComponentParameterNode) -> ExpressionNode:
        structure = self.structure_provider.get_component_parameter_structure(
            node.component_id, node.name
        )
        time_index = TimeShift(0) if structure.time else NoTimeIndex()
        scenario_index = (
            CurrentScenarioIndex() if structure.scenario else NoScenarioIndex()
        )
        return problem_param(node.component_id, node.name, time_index, scenario_index)

    def time_shift(self, node: TimeShiftNode) -> ExpressionNode:
        shift = self.evaluator(node.time_shift)
        operand = visit(node.operand, self)
        return apply_timeshift(operand, shift)

    def time_eval(self, node: TimeEvalNode) -> ExpressionNode:
        timestep = self.evaluator(node.eval_time)
        operand = visit(node.operand, self)
        return apply_timestep(operand, timestep)

    def time_sum(self, node: TimeSumNode) -> ExpressionNode:
        from_shift = self.evaluator(node.from_time)
        to_shift = self.evaluator(node.to_time)
        operand = visit(node.operand, self)
        nodes = []
        for t in range(from_shift, to_shift + 1):
            nodes.append(apply_timeshift(operand, t))
        return sum_expressions(nodes)

    def all_time_sum(self, node: AllTimeSumNode) -> ExpressionNode:
        nodes = []
        operand = visit(node.operand, self)
        for t in range(self.timesteps_count):
            # if we sum previously "evaluated" variables for example x[0], it's ok
            nodes.append(apply_timestep(operand, t, allow_existing=True))
        return sum_expressions(nodes)

    def scenario_operator(self, node: ScenarioOperatorNode) -> ExpressionNode:
        if node.name != "Expectation":
            raise ValueError(f"Scenario operator not supported: {node.name}")
        nodes = []
        operand = visit(node.operand, self)
        for t in range(self.scenarios_count):
            nodes.append(apply_scenario(operand, t))
        return sum_expressions(nodes) / self.scenarios_count

    def pb_parameter(self, node: ProblemParameterNode) -> ExpressionNode:
        raise ValueError("Should not reach")

    def pb_variable(self, node: ProblemVariableNode) -> ExpressionNode:
        raise ValueError("Should not reach")


def expand_operators(
    expression: ExpressionNode,
    dimensions: ProblemDimensions,
    evaluator: ExpressionEvaluator,
    structure_provider: IndexingStructureProvider,
) -> ExpressionNode:
    return visit(
        expression,
        OperatorsExpansion(
            dimensions.timesteps_count,
            dimensions.scenarios_count,
            evaluator,
            structure_provider,
        ),
    )


TimeIndexedNode = TypeVar(
    "TimeIndexedNode", bound=Union[ProblemParameterNode, ProblemVariableNode]
)


@dataclass(frozen=True)
class ApplyTimeShift(CopyVisitor):
    """
    Shifts all underlying expressions.
    """

    timeshift: int

    def _apply_timeshift(self, node: TimeIndexedNode) -> TimeIndexedNode:
        current_index = node.time_index
        if isinstance(current_index, TimeShift):
            return dataclasses.replace(
                node, time_index=TimeShift(current_index.timeshift + self.timeshift)
            )
        if isinstance(current_index, TimeStep):
            return dataclasses.replace(
                node, time_index=TimeStep(current_index.timestep + self.timeshift)
            )
        if isinstance(current_index, NoTimeIndex):
            return node
        raise ValueError("Unknown time index type.")

    def pb_parameter(self, node: ProblemParameterNode) -> ProblemParameterNode:
        return self._apply_timeshift(node)

    def pb_variable(self, node: ProblemVariableNode) -> ProblemVariableNode:
        return self._apply_timeshift(node)


def apply_timeshift(expression: ExpressionNode, timeshift: int) -> ExpressionNode:
    return visit(expression, ApplyTimeShift(timeshift))


@dataclass(frozen=True)
class ApplyTimeStep(CopyVisitor):
    """
    Applies timestep to all underlying expressions.
    """

    timestep: int
    allow_existing: bool = False

    def _apply_timestep(self, node: TimeIndexedNode) -> TimeIndexedNode:
        current_index = node.time_index
        if isinstance(current_index, TimeShift):
            return dataclasses.replace(
                node, time_index=TimeStep(current_index.timeshift + self.timestep)
            )
        if isinstance(current_index, TimeStep):
            if not self.allow_existing:
                raise ValueError(
                    "Cannot override a previously defined timestep (for example (x[0])[1])."
                )
            return node
        if isinstance(current_index, NoTimeIndex):
            return node
        raise ValueError("Unknown time index type.")

    def pb_parameter(self, node: ProblemParameterNode) -> ExpressionNode:
        return self._apply_timestep(node)

    def pb_variable(self, node: ProblemVariableNode) -> ExpressionNode:
        return self._apply_timestep(node)


def apply_timestep(
    expression: ExpressionNode, timestep: int, allow_existing: bool = False
) -> ExpressionNode:
    return visit(expression, ApplyTimeStep(timestep, allow_existing))


@dataclass(frozen=True)
class ApplyScenario(CopyVisitor):
    scenario: int

    def pb_parameter(self, node: ProblemParameterNode) -> ExpressionNode:
        if isinstance(node.scenario_index, NoScenarioIndex):
            return node
        return dataclasses.replace(node, scenario_index=OneScenarioIndex(self.scenario))

    def pb_variable(self, node: ProblemVariableNode) -> ExpressionNode:
        if isinstance(node.scenario_index, NoScenarioIndex):
            return node
        return dataclasses.replace(node, scenario_index=OneScenarioIndex(self.scenario))


def apply_scenario(expression: ExpressionNode, scenario: int) -> ExpressionNode:
    return visit(expression, ApplyScenario(scenario))



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

"""
The network module defines the data model for an instance of network,
including nodes, links, and components (model instantations).
"""
import itertools
from dataclasses import dataclass, field, replace
from typing import Any, Dict, Iterable, List, cast

from gems.model import PortField, PortType
from gems.model.model import Model
from gems.model.port import PortFieldId
from gems.utils import require_not_none


@dataclass(frozen=True)
class Component:
    """
    A component is an instance of a model, with specified parameter values.
    """

    model: Model
    id: str

    def is_variable_in_model(self, var_id: str) -> bool:
        return var_id in self.model.variables.keys()

    def replicate(self, /, **changes: Any) -> "Component":
        return replace(self, **changes)


def create_component(model: Model, id: str) -> Component:
    return Component(model=model, id=id)


@dataclass(frozen=True)
class Node(Component):
    """
    A node in the network.
    """

    pass


def create_node(model: Model, id: str) -> Node:
    return Node(model=model, id=id)


@dataclass(frozen=True)
class PortRef:
    component: Component
    port_id: str


@dataclass()
class PortsConnection:
    port1: PortRef
    port2: PortRef
    master_port: Dict[PortField, PortRef] = field(init=False, default_factory=dict)

    def __post_init__(self) -> None:
        self.__validate_ports()

    def __validate_ports(self) -> None:
        model1 = self.port1.component.model
        model2 = self.port2.component.model
        port_1 = model1.ports.get(self.port1.port_id)
        port_2 = model2.ports.get(self.port2.port_id)

        if port_1 is None or port_2 is None:
            raise ValueError(f"Missing port: {port_1} or {port_2} ")
        if port_1.port_type != port_2.port_type:
            raise ValueError(
                f"Incompatible portTypes {port_1.port_type} != {port_2.port_type}"
            )

        for field_name in [f.name for f in port_1.port_type.fields]:
            def1: bool = (
                PortFieldId(port_name=port_1.port_name, field_name=field_name)
                in model1.port_fields_definitions
            )
            def2: bool = (
                PortFieldId(port_name=port_2.port_name, field_name=field_name)
                in model2.port_fields_definitions
            )
            if not def1 and not def2:
                raise ValueError(
                    f"No definition for port field {field_name} on {port_1.port_name}."
                )
            if def1 and def2:
                raise ValueError(
                    f"Port field {field_name} on {port_1.port_name} has 2 definitions."
                )

            self.master_port[PortField(name=field_name)] = (
                self.port1 if def1 else self.port2
            )

    def get_port_type(self) -> PortType:
        port_1 = self.port1.component.model.ports.get(self.port1.port_id)

        if port_1 is None:
            raise ValueError(f"Missing port: {port_1}")
        return port_1.port_type

    def replicate(self, /, **changes: Any) -> "PortsConnection":
        # Shallow copy
        return replace(self, **changes)


@dataclass
class Network:
    """
    Network model: simply nodes, links, and components.
    """

    id: str
    _nodes: Dict[str, Node] = field(init=False, default_factory=dict)
    _components: Dict[str, Component] = field(init=False, default_factory=dict)
    _connections: List[PortsConnection] = field(init=False, default_factory=list)

    def _check_node_exists(self, node_id: str) -> None:
        if node_id not in self._nodes:
            raise ValueError(f"Node {node_id} does not exist in the network.")

    def add_component(self, component: Component) -> None:
        require_not_none(component)
        self._components[component.id] = component

    def get_component(self, component_id: str) -> Component:
        """
        Returns the component (possibly a node) corresponding to this ID.
        """
        res = self._components.get(component_id, None)
        return res if res else self._nodes[component_id]

    @property
    def components(self) -> Iterable[Component]:
        return self._components.values()

    def add_node(self, node: Node) -> None:
        self._nodes[node.id] = node

    def get_node(self, node_id: str) -> Node:
        return self._nodes[node_id]

    @property
    def nodes(self) -> Iterable[Node]:
        return self._nodes.values()

    @property
    def all_components(self) -> Iterable[Component]:
        """
        An iterable over both nodes and components.
        """
        return itertools.chain(self.nodes, self.components)

    def connect(self, port1: PortRef, port2: PortRef) -> None:
        ports_connection = PortsConnection(port1, port2)
        self._connections.append(ports_connection)

    @property
    def connections(self) -> Iterable[PortsConnection]:
        return self._connections

    def get_connection(self, idx: int) -> PortsConnection:
        return self._connections[idx]

    def is_empty(self) -> bool:
        return (not self._nodes) and (not self._components) and (not self._connections)

    def replicate(self, /, **changes: Any) -> "Network":
        replica = replace(self, **changes)

        for node in self.nodes:
            replica.add_node(cast(Node, node.replicate()))

        for component in self.components:
            replica.add_component(component.replicate())

        for connection in self.connections:
            replica._connections.append(connection.replicate())

        return replica



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.
from abc import ABC, abstractmethod
from dataclasses import dataclass
from pathlib import Path
from typing import Dict, Mapping, Optional

import pandas as pd

from gems.study.network import Network


@dataclass(frozen=True)
class TimeScenarioIndex:
    time: int
    scenario: int


@dataclass(frozen=True)
class TimeIndex:
    time: int


@dataclass(frozen=True)
class ScenarioIndex:
    scenario: int


@dataclass(frozen=True)
class Scenarization:
    _scenarization: Dict[int, int]

    def get_scenario_for_year(self, year: int) -> int:
        return self._scenarization[year]

    def add_year(self, year: int, scenario: int) -> None:
        if year in self._scenarization:
            raise ValueError(f"the year {year} is already defined")
        self._scenarization[year] = scenario


@dataclass(frozen=True)
class AbstractDataStructure(ABC):
    @abstractmethod
    def get_value(
        self, timestep: Optional[int], scenario: Optional[int], node_id: str = ""
    ) -> float:
        raise NotImplementedError()

    @abstractmethod
    def check_requirement(self, time: bool, scenario: bool) -> bool:
        """
        Check if the data structure meets certain requirements.
        Implement this method in subclasses as needed.
        """
        pass


@dataclass(frozen=True)
class ConstantData(AbstractDataStructure):
    value: float

    def get_value(
        self, timestep: Optional[int], scenario: Optional[int], node_id: str = ""
    ) -> float:
        return self.value

    # ConstantData can be used for time varying or constant models
    def check_requirement(self, time: bool, scenario: bool) -> bool:
        if not isinstance(self, ConstantData):
            raise ValueError("Invalid data type for ConstantData")
        return True


@dataclass(frozen=True)
class TimeSeriesData(AbstractDataStructure):
    """
    Container for identifiable timeseries data.
    When a model is instantiated as a component, property values
    can be defined by referencing one of those timeseries by its ID.
    """

    time_series: Mapping[TimeIndex, float]

    def get_value(
        self, timestep: Optional[int], scenario: Optional[int], node_id: str = ""
    ) -> float:
        if timestep is None:
            raise KeyError("Time series data requires a time index.")
        return self.time_series[TimeIndex(timestep)]

    def check_requirement(self, time: bool, scenario: bool) -> bool:
        if not isinstance(self, TimeSeriesData):
            raise ValueError("Invalid data type for TimeSeriesData")

        return time


@dataclass(frozen=True)
class ScenarioSeriesData(AbstractDataStructure):
    """
    Container for identifiable timeseries data.
    When a model is instantiated as a component, property values
    can be defined by referencing one of those timeseries by its ID.
    """

    scenario_series: Mapping[ScenarioIndex, float]
    scenarization: Optional[Scenarization] = None

    def get_value(
        self, timestep: Optional[int], scenario: Optional[int], node_id: str = ""
    ) -> float:
        if scenario is None:
            raise KeyError("Scenario series data requires a scenario index.")
        if self.scenarization:
            scenario = self.scenarization.get_scenario_for_year(scenario)
        return self.scenario_series[ScenarioIndex(scenario)]

    def check_requirement(self, time: bool, scenario: bool) -> bool:
        if not isinstance(self, ScenarioSeriesData):
            raise ValueError("Invalid data type for TimeSeriesData")

        return scenario


def load_ts_from_file(
    timeseries_name: Optional[str], path_to_file: Optional[Path]
) -> pd.DataFrame:
    if path_to_file is None or timeseries_name is None:
        raise FileNotFoundError(f"File '{timeseries_name}' does not exist")

    base_path = path_to_file / timeseries_name
    candidates = [base_path.with_suffix(".txt"), base_path.with_suffix(".tsv")]

    last_exc: Optional[Exception] = None
    for candidate in candidates:
        if not candidate.exists():
            continue
        try:
            sep = r"\s+" if candidate.suffix == ".txt" else "\t"
            return pd.read_csv(candidate, header=None, sep=sep)
        except Exception as e:
            last_exc = e
            break

    if last_exc is not None:
        raise Exception(
            f"An error has arrived when processing '{candidate}': {last_exc}"
        )

    raise FileNotFoundError(
        f"File '{timeseries_name}.txt' or '{timeseries_name}.tsv' does not exist"
    )


def dataframe_to_time_series(ts_dataframe: pd.DataFrame) -> Dict[TimeIndex, float]:
    if ts_dataframe.shape[1] != 1:
        raise ValueError(
            f"Could not convert input data to time series data. Expect data series with exactly one column, got shape {ts_dataframe.shape}"
        )
    df_index = ts_dataframe.index.astype(int)  # Only for mypy
    return {
        TimeIndex(index): float(value)
        for index, value in zip(df_index, ts_dataframe.iloc[:, 0].values)
    }


def dataframe_to_scenario_series(
    ts_dataframe: pd.DataFrame,
) -> Dict[ScenarioIndex, float]:
    if ts_dataframe.shape[0] != 1:
        raise ValueError(
            f"Could not convert input data to scenario series data. Expect data series with exactly one line, got shape {ts_dataframe.shape}"
        )

    return {
        ScenarioIndex(col_id): float(value)
        for col_id, value in zip(
            list(range(ts_dataframe.shape[1])), ts_dataframe.iloc[0, :].values
        )
    }


@dataclass(frozen=True)
class TimeScenarioSeriesData(AbstractDataStructure):
    """
    Container for identifiable timeseries data.
    When a model is instantiated as a component, property values
    can be defined by referencing one of those timeseries by its ID.
    """

    time_scenario_series: pd.DataFrame
    scenarization: Optional[Scenarization] = None

    def get_value(
        self, timestep: Optional[int], scenario: Optional[int], node_id: str = ""
    ) -> float:
        if timestep is None:
            raise KeyError("Time scenario data requires a time index.")
        if scenario is None:
            raise KeyError("Time scenario data requires a scenario index.")
        if self.scenarization:
            scenario = self.scenarization.get_scenario_for_year(scenario)
        value = str(self.time_scenario_series.iloc[timestep, scenario])
        return float(value)

    def check_requirement(self, time: bool, scenario: bool) -> bool:
        if not isinstance(self, TimeScenarioSeriesData):
            raise ValueError("Invalid data type for TimeScenarioSeriesData")

        return time and scenario


@dataclass(frozen=True)
class TreeData(AbstractDataStructure):
    data: Mapping[str, AbstractDataStructure]

    def get_value(
        self, timestep: Optional[int], scenario: Optional[int], node_id: str = ""
    ) -> float:
        return self.data[node_id].get_value(timestep, scenario)

    def check_requirement(self, time: bool, scenario: bool) -> bool:
        return all(
            node_data.check_requirement(time, scenario)
            for node_data in self.data.values()
        )


@dataclass(frozen=True)
class ComponentParameterIndex:
    component_id: str
    parameter_name: str


class DataBase:
    """
    Container for identifiable data.
    When a model is instantiated as a component, property values
    can be defined by referencing one of those data by its ID.
    Data can have different structure : constant, varying in time or scenarios.
    """

    _data: Dict[ComponentParameterIndex, AbstractDataStructure]

    def __init__(self) -> None:
        self._data: Dict[ComponentParameterIndex, AbstractDataStructure] = {}

    def get_data(self, component_id: str, parameter_name: str) -> AbstractDataStructure:
        return self._data[ComponentParameterIndex(component_id, parameter_name)]

    def add_data(
        self, component_id: str, parameter_name: str, data: AbstractDataStructure
    ) -> None:
        self._data[ComponentParameterIndex(component_id, parameter_name)] = data

    def get_value(
        self, index: ComponentParameterIndex, timestep: int, scenario: int
    ) -> float:
        if index in self._data:
            return self._data[index].get_value(timestep, scenario)
        else:
            raise KeyError(f"Index {index} not found.")

    def requirements_consistency(self, network: Network) -> None:
        for component in network.components:
            for param in component.model.parameters.values():
                data_structure = self.get_data(component.id, param.name)

                if not data_structure.check_requirement(
                    component.model.parameters[param.name].structure.time,
                    component.model.parameters[param.name].structure.scenario,
                ):
                    raise ValueError(
                        f"Data inconsistency for component: {component.id}, parameter: {param.name}. Requirement not met."
                    )



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from dataclasses import dataclass
from typing import List, Optional


# TODO: Move keys elsewhere as variables have no sense in this file
@dataclass(eq=True, frozen=True)
class TimestepComponentVariableKey:
    """
    Identifies the solver variable for one timestep and one component variable.
    """

    tree_node_name: str
    component_id: str
    variable_name: str
    block_timestep: Optional[int] = None
    scenario: Optional[int] = None


@dataclass(frozen=True)
class TimeBlock:
    """
    One block for otimization (week in current tool).

    timesteps: list of the different timesteps of the block (0, 1, ... 168 for each hour in one week)
    """

    id: int
    timesteps: List[int]



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from abc import ABC, abstractmethod
from dataclasses import dataclass

from gems.expression.evaluate import ValueProvider

from .copy import CopyVisitor
from .expression import (
    ComponentParameterNode,
    ExpressionNode,
    LiteralNode,
    ParameterNode,
)
from .visitor import visit


class ParameterValueProvider(ABC):
    @abstractmethod
    def get_parameter_value(self, name: str) -> float:
        ...

    @abstractmethod
    def get_component_parameter_value(self, component_id: str, name: str) -> float:
        ...


@dataclass(frozen=True)
class ParameterResolver(CopyVisitor):
    """
    Duplicates the AST with replacement of parameter nodes by literal nodes.
    """

    context: ParameterValueProvider

    def parameter(self, node: ParameterNode) -> ExpressionNode:
        value: float = self.context.get_parameter_value(node.name)
        return LiteralNode(value)

    def comp_parameter(self, node: ComponentParameterNode) -> ExpressionNode:
        value: float = self.context.get_component_parameter_value(
            node.component_id, node.name
        )
        return LiteralNode(value)


def resolve_parameters(
    expression: ExpressionNode, parameter_provider: ParameterValueProvider
) -> ExpressionNode:
    return visit(expression, ParameterResolver(parameter_provider))



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.
import logging
import os
import sys
from argparse import ArgumentParser, ArgumentTypeError, Namespace
from configparser import ConfigParser
from pathlib import Path

from gems.input_converter.src.data_preprocessing.data_classes import ConversionMode

from .converter import AntaresStudyConverter
from .logger import Logger

DEFAULT: dict = {}
LOGGER_PATH: str = os.path.join(
    os.path.dirname(os.path.dirname(__file__)), "data/logging.log"
)


class CreateFileIfMissing:
    """
    Argparse type that accepts a file path and creates the file if it doesn't exist.
    Fails if the path points to an existing directory or if the parent directory is missing.
    """

    def __call__(self, string: str) -> Path:
        path = Path(string).expanduser().resolve()

        if path.exists() and not path.is_file():
            raise ArgumentTypeError(f"Path exists and is not a file: '{path}'")

        if not path.parent.exists():
            raise ArgumentTypeError(f"Parent directory does not exist: '{path.parent}'")

        if not path.exists():
            try:
                path.touch()
            except Exception as e:
                raise ArgumentTypeError(f"Failed to create file: {e}")

        return path


class PathType:
    def __init__(
        self,
        exists: bool = False,
        file_ok: bool = False,
        dir_ok: bool = False,
    ) -> None:
        if not (file_ok or dir_ok):
            msg = "Either `file_ok` or `dir_ok` must be set at a minimum."
            raise ValueError(msg)
        self.exists = exists
        self.file_ok = file_ok
        self.dir_ok = dir_ok

    def __call__(self, string: str) -> Path:
        file_path = Path(string).expanduser()
        if not self.exists:
            return file_path
        if self.file_ok and self.dir_ok:
            if file_path.exists():
                return file_path
            msg = f"The file or directory path does not exist: '{file_path}'"
            raise ArgumentTypeError(msg)
        elif self.file_ok:
            if file_path.is_file():
                return file_path
            elif file_path.exists():
                msg = f"The path is not a regular file: '{file_path}'"
            else:
                msg = f"The file path does not exist: '{file_path}'"
            raise ArgumentTypeError(msg)
        elif self.dir_ok:
            if file_path.is_dir():
                return file_path
            elif file_path.exists():
                msg = f"The path is not a directory: '{file_path}'"
            else:
                msg = f"The directory path does not exist: '{file_path}'"
            raise ArgumentTypeError(msg)
        else:  # pragma: no cover
            raise NotImplementedError((self.file_ok, self.dir_ok))


def parse_commandline() -> Namespace:
    """Parse command-line arguments using argparse to specify configuration file paths,
    logging options, and version display. Returns the parsed arguments."""
    parser = ArgumentParser()
    parser.add_argument(
        "-c",
        "--conf",
        type=PathType(exists=True, file_ok=True),
        help="Give the path of the configuration file",
        default="data/config.ini",
    )
    parser.add_argument(
        "-l",
        "--logging",
        type=CreateFileIfMissing(),
        help="Give the path of the logger file",
        default=LOGGER_PATH,
    )
    return parser.parse_args()


if __name__ == "__main__":
    config: dict = {}
    args = parse_commandline()
    log_path = args.logging
    log_path.parent.mkdir(parents=True, exist_ok=True)

    logger: logging.Logger = Logger(__name__, log_path)

    config_parser = ConfigParser()

    # Load the default configuration dictionary into the config parser.
    config_parser.read_dict(DEFAULT)

    if args.conf:
        config_parser.read(args.conf)

    if not config_parser.has_section("study"):
        sys.exit(f"Aborting: missing 'study' section in configuration")

    try:
        path_validator = PathType(exists=True, dir_ok=True)
        study_path = path_validator(config_parser.get("study", "study_path"))
        output_folder = path_validator(config_parser.get("study", "output_folder"))
        lib_paths = list(
            filter(None, config_parser.get("study", "lib_paths").split(","))
        )
        try:
            for path in lib_paths:
                PathType(exists=True, file_ok=True)(path)
        except ArgumentTypeError as e:
            sys.exit(f"Aborting: Invalid lib_paths in configuration file: {e}")
    except ArgumentTypeError as e:
        sys.exit(f"Aborting: Invalid path in configuration file: {e}")

    mode = config_parser.get("study", "mode", fallback=ConversionMode.FULL.value)
    params = {
        "study_input": study_path,
        "logger": logger,
        "mode": mode,
        "output_folder": output_folder,
        "lib_paths": lib_paths,
    }
    if mode == ConversionMode.HYBRID.value:
        if not config_parser.get("hybrid", "model_list"):
            sys.exit(f"Aborting: 'hybrid' mode set but no config related")

        model_list = [
            model.strip()
            for model in config_parser.get("hybrid", "model_list").split(",")
            if model.strip()
        ]

        params["model_list"] = model_list

    converter = AntaresStudyConverter(**params)  # type: ignore
    converter.process_all()



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

import argparse
import os
from dataclasses import dataclass
from pathlib import Path
from typing import List, TextIO

import pandas as pd
from yaml import safe_load

from gems.pypsa_converter.models.gems_system_yaml_schema import GemsSystem


def parse_yaml_components(input_study: TextIO) -> GemsSystem:
    tree = safe_load(input_study)
    return GemsSystem.model_validate(tree["system"])


def parse_scenario_builder(file: Path) -> pd.DataFrame:
    sb = pd.read_csv(file, names=("name", "year", "scenario"))
    sb.rename(columns={0: "name", 1: "year", 2: "scenario"})
    return sb


@dataclass(frozen=True)
class ParsedArguments:
    models_path: List[Path]
    components_path: Path
    timeseries_path: Path
    duration: int
    nb_scenarios: int


def parse_cli() -> ParsedArguments:
    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--study", type=Path, help="path to the root directory of the study"
    )
    parser.add_argument(
        "--models", nargs="+", type=Path, help="list of path to model file, *.yml"
    )
    parser.add_argument(
        "--component", type=Path, help="path to the component file, *.yml"
    )
    parser.add_argument(
        "--timeseries", type=Path, help="path to the timeseries directory"
    )
    parser.add_argument(
        "--duration", type=int, help="duration of the simulation", default=1
    )
    parser.add_argument(
        "--scenario", type=int, help="number of scenario of the simulation", default=1
    )

    args = parser.parse_args()

    if args.study:
        if args.models or args.component or args.timeseries:
            parser.error(
                "--study flag can't be use with --models, --component and --timeseries"
            )

        components_path = args.study / "input" / "components" / "components.yml"
        timeseries_dir = args.study / "input" / "components" / "series"
        model_paths = [
            args.study / "input" / "models" / file
            for file in os.listdir(args.study / "input" / "models")
        ]

    else:
        if not args.models or not args.component:
            parser.error("--models and --component must be entered")

        components_path = args.component
        timeseries_dir = args.timeseries
        model_paths = args.models

    return ParsedArguments(
        model_paths, components_path, timeseries_dir, args.duration, args.scenario
    )



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from dataclasses import dataclass
from typing import Dict, List

from gems.expression import CopyVisitor, sum_expressions, visit
from gems.expression.expression import (
    ExpressionNode,
    PortFieldAggregatorNode,
    PortFieldNode,
)
from gems.model.port import PortFieldId


@dataclass(eq=True, frozen=True)
class PortFieldKey:
    """
    Identifies the expression node for one component and one port variable.
    """

    component_id: str
    port_variable_id: PortFieldId


@dataclass(frozen=True)
class PortResolver(CopyVisitor):
    """
    Duplicates the AST with replacement of port field nodes by
    their corresponding expression.
    """

    component_id: str
    ports_expressions: Dict[PortFieldKey, List[ExpressionNode]]

    def port_field(self, node: PortFieldNode) -> ExpressionNode:
        expressions = self.ports_expressions[
            PortFieldKey(
                self.component_id, PortFieldId(node.port_name, node.field_name)
            )
        ]
        if len(expressions) != 1:
            raise ValueError(
                f"Invalid number of expression for port : {node.port_name}"
            )
        else:
            return expressions[0]

    def port_field_aggregator(self, node: PortFieldAggregatorNode) -> ExpressionNode:
        if node.aggregator != "PortSum":
            raise NotImplementedError("Only PortSum is supported.")
        port_field_node = node.operand
        if not isinstance(port_field_node, PortFieldNode):
            raise ValueError(f"Should be a portFieldNode : {port_field_node}")

        expressions = self.ports_expressions.get(
            PortFieldKey(
                self.component_id,
                PortFieldId(port_field_node.port_name, port_field_node.field_name),
            ),
            [],
        )
        return sum_expressions(expressions)


def resolve_port(
    expression: ExpressionNode,
    component_id: str,
    ports_expressions: Dict[PortFieldKey, List[ExpressionNode]],
) -> ExpressionNode:
    return visit(expression, PortResolver(component_id, ports_expressions))



import os
from enum import Enum
from pathlib import Path
from typing import Callable

import numpy as np
import pandas as pd
from antares.craft.model.thermal import ThermalCluster

from gems.study.parsing import InputComponentParameter


class Direction(Enum):
    FORWARD = "forward"
    BACKWARD = "backward"


class ThermalDataPreprocessing:
    DEFAULT_PERIOD: int = 168

    def __init__(self, thermal: ThermalCluster, study_path: Path, suffix: str = ".tsv"):
        self.thermal = thermal
        self.study_path = study_path
        self.suffix = suffix
        self.output_series_dir = self.study_path / "input" / "data-series"
        self._prepro_parameter_functions: dict[str, Callable[[int], pd.DataFrame]] = {
            "p_min_cluster": lambda _: self._compute_p_min_cluster(),
            "p_max_cluster": lambda _: self._compute_p_max_cluster(),
            "nb_units_min": lambda _: self._compute_nb_units_min(),
            "nb_units_max": lambda _: self._compute_nb_units_max(),
            "nb_units_max_variation_forward": lambda period: self._compute_nb_units_max_variation(
                Direction.FORWARD, period
            ),
            "nb_units_max_variation_backward": lambda period: self._compute_nb_units_max_variation(
                Direction.BACKWARD, period
            ),
        }

    def _compute_p_min_cluster(self) -> pd.DataFrame:
        modulation_data: pd.Series = self.thermal.get_prepro_modulation_matrix().iloc[
            :, 3
        ]
        series_data: pd.DataFrame = self.thermal.get_series_matrix()
        unit_count: int = self.thermal.properties.unit_count
        nominal_capacity: float = self.thermal.properties.nominal_capacity
        scaled_modulation: pd.Series = modulation_data * nominal_capacity * unit_count
        #  min(min_gen_modulation * unit_count * nominal_capacity, p_max_cluster)
        min_values: pd.Series = pd.concat([scaled_modulation, series_data], axis=1).min(
            axis=1
        )
        return min_values.to_frame(name="p_min_cluster")

    def _compute_p_max_cluster(self) -> pd.DataFrame:
        return self.thermal.get_series_matrix()

    def _compute_nb_units_min(self) -> pd.DataFrame:
        p_min_cluster: pd.DataFrame = self._compute_p_min_cluster()
        nominal_capacity: float = self.thermal.properties.nominal_capacity
        return pd.DataFrame(
            np.ceil(p_min_cluster / nominal_capacity),
        )

    def _compute_nb_units_max(self) -> pd.DataFrame:
        series_data: pd.DataFrame = self.thermal.get_series_matrix()
        nominal_capacity: float = self.thermal.properties.nominal_capacity
        return pd.DataFrame(
            np.ceil(series_data / nominal_capacity),
        )

    def _compute_nb_units_max_variation(
        self, direction: Direction, period: int = DEFAULT_PERIOD
    ) -> pd.DataFrame:
        nb_units_max = self._compute_nb_units_max()
        indices = np.arange(len(nb_units_max))
        max_valid_index = len(nb_units_max) - 1  # 8759

        previous_indices: np.ndarray = (indices - 1) % period + (
            indices // period
        ) * period

        previous_indices = np.where(
            indices == 0, 0, np.minimum(previous_indices, max_valid_index)
        )
        variation = pd.DataFrame()
        if direction == Direction.BACKWARD:
            variation = nb_units_max.reset_index(drop=True) - nb_units_max.iloc[
                previous_indices
            ].reset_index(drop=True)
        elif direction == Direction.FORWARD:
            variation = nb_units_max.iloc[previous_indices].reset_index(
                drop=True
            ) - nb_units_max.reset_index(drop=True)

        # Usage of vectorized operation instead of applymap
        # It is the equivalent of max(0, variation(x))
        variation = variation.clip(lower=0)
        return variation.rename(
            columns={variation.columns[0]: f"nb_units_max_variation_{direction.value}"}
        )

    def _build_csv_path_and_name(self, param_id: str) -> tuple[Path, str]:
        name = f"{self.thermal.area_id}_{self.thermal.id}_{param_id}"
        return self.output_series_dir / str(name + self.suffix), name

    def generate_component_parameter(
        self, parameter_id: str, period: int = 0
    ) -> InputComponentParameter:
        if parameter_id not in self._prepro_parameter_functions:
            raise ValueError(f"Unsupported parameter_id: {parameter_id}")

        df = self._prepro_parameter_functions[parameter_id](period)
        csv_path, value_name = self._build_csv_path_and_name(parameter_id)

        output_dir = os.path.dirname(csv_path)
        if output_dir:
            os.makedirs(output_dir, exist_ok=True)
        # This separator is chosen to comply with the antares_craft timeseries creation
        df.to_csv(csv_path, sep="\t", index=False, header=False)

        return InputComponentParameter(
            id=parameter_id,
            time_dependent=True,
            scenario_dependent=True,
            value=value_name,
        )



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

"""
Operators that allow port manipulation of expressions
"""

from abc import ABC
from dataclasses import dataclass


@dataclass(frozen=True)
class PortOperator(ABC):
    pass


@dataclass(frozen=True)
class PortAggregator:
    pass


@dataclass(frozen=True)
class PortSum(PortAggregator):
    pass



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

"""
Operators that allow manipulation of expressions with respect to scenarios
"""

from abc import ABC, abstractmethod
from dataclasses import dataclass


@dataclass(frozen=True)
class ScenarioOperator(ABC):
    def __str__(self) -> str:
        return NotImplemented

    @classmethod
    @abstractmethod
    def degree(cls) -> int:
        raise NotImplementedError


@dataclass(frozen=True)
class Expectation(ScenarioOperator):
    def __str__(self) -> str:
        return "expec()"

    @classmethod
    def degree(cls) -> int:
        return 1


@dataclass(frozen=True)
class Variance(ScenarioOperator):
    def __str__(self) -> str:
        return "variance()"

    @classmethod
    def degree(cls) -> int:
        return 2



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

import gems.expression.scenario_operator
from gems.expression.expression import (
    AllTimeSumNode,
    ComponentParameterNode,
    ComponentVariableNode,
    PortFieldAggregatorNode,
    PortFieldNode,
    ProblemParameterNode,
    ProblemVariableNode,
    TimeEvalNode,
    TimeShiftNode,
    TimeSumNode,
)

from .expression import (
    AdditionNode,
    ComparisonNode,
    DivisionNode,
    ExpressionNode,
    LiteralNode,
    MultiplicationNode,
    NegationNode,
    ParameterNode,
    ScenarioOperatorNode,
    VariableNode,
)
from .visitor import ExpressionVisitor, T, visit


class ExpressionDegreeVisitor(ExpressionVisitor[int]):
    """
    Computes degree of expression with respect to variables.
    """

    def literal(self, node: LiteralNode) -> int:
        return 0

    def negation(self, node: NegationNode) -> int:
        return visit(node.operand, self)

    # TODO: Take into account simplification that can occur with literal coefficient for add, sub, mult, div
    def addition(self, node: AdditionNode) -> int:
        degrees = [visit(o, self) for o in node.operands]
        return max(degrees)

    def multiplication(self, node: MultiplicationNode) -> int:
        return visit(node.left, self) + visit(node.right, self)

    def division(self, node: DivisionNode) -> int:
        right_degree = visit(node.right, self)
        if right_degree != 0:
            raise ValueError("Degree computation not implemented for divisions.")
        return visit(node.left, self)

    def comparison(self, node: ComparisonNode) -> int:
        return max(visit(node.left, self), visit(node.right, self))

    def variable(self, node: VariableNode) -> int:
        return 1

    def parameter(self, node: ParameterNode) -> int:
        return 0

    def comp_variable(self, node: ComponentVariableNode) -> int:
        return 1

    def comp_parameter(self, node: ComponentParameterNode) -> int:
        return 0

    def pb_variable(self, node: ProblemVariableNode) -> int:
        return 1

    def pb_parameter(self, node: ProblemParameterNode) -> int:
        return 0

    def time_shift(self, node: TimeShiftNode) -> int:
        return visit(node.operand, self)

    def time_eval(self, node: TimeEvalNode) -> int:
        return visit(node.operand, self)

    def time_sum(self, node: TimeSumNode) -> int:
        return visit(node.operand, self)

    def all_time_sum(self, node: AllTimeSumNode) -> int:
        return visit(node.operand, self)

    def scenario_operator(self, node: ScenarioOperatorNode) -> int:
        scenario_operator_cls = getattr(gems.expression.scenario_operator, node.name)
        # TODO: Carefully check if this formula is correct
        return scenario_operator_cls.degree() * visit(node.operand, self)

    def port_field(self, node: PortFieldNode) -> int:
        return 1

    def port_field_aggregator(self, node: PortFieldAggregatorNode) -> int:
        return visit(node.operand, self)


def compute_degree(expression: ExpressionNode) -> int:
    return visit(expression, ExpressionDegreeVisitor())


def is_constant(expr: ExpressionNode) -> bool:
    """
    True if the expression has no variable.
    """
    return compute_degree(expr) == 0


def is_linear(expr: ExpressionNode) -> bool:
    """
    True if the expression is linear with respect to variables.
    """
    return compute_degree(expr) <= 1



from dataclasses import dataclass


@dataclass
class PyPSAGlobalConstraintData:
    pypsa_name: str
    # pypsa_investment_period
    pypsa_carrier_attribute: str
    pypsa_sense: str
    pypsa_constant: float
    gems_model_id: str  # gems model for this GlobalConstraint
    gems_port_id: str  # gems port for this GlobalConstraint
    gems_components_and_ports: list[tuple[str, str]]



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.
import logging
import sys
from typing import Optional


def Logger(name: str, file_name: Optional[str]) -> logging.Logger:
    formatter = logging.Formatter(
        fmt="%(asctime)s %(module)s,line: %(lineno)d %(levelname)8s | %(message)s",
        datefmt="%Y/%m/%d %H:%M:%S",
    )

    if file_name and not str(file_name).endswith(".log"):
        file_name = f"{file_name}.log"
    logging.basicConfig(
        filename=file_name,
        format="%(asctime)s %(module)s,line: %(lineno)d %(levelname)8s | %(message)s",
        datefmt="%Y/%m/%d %H:%M:%S",
        filemode="a",
        level=logging.INFO,
    )

    log_obj = logging.getLogger(name)
    log_obj.setLevel(logging.DEBUG)

    # console printer
    screen_handler = logging.StreamHandler(stream=sys.stdout)
    screen_handler.setFormatter(formatter)
    log_obj.addHandler(screen_handler)

    log_obj.info("Logger object created successfully.. ")
    return log_obj



# Copyright (c) 2024, RTE (https://www.rte-france.com)
#
# See AUTHORS.txt
#
# This Source Code Form is subject to the terms of the Mozilla Public
# License, v. 2.0. If a copy of the MPL was not distributed with this
# file, You can obtain one at http://mozilla.org/MPL/2.0/.
#
# SPDX-License-Identifier: MPL-2.0
#
# This file is part of the Antares project.

from typing import Any

import numpy as np
import yaml
from pydantic import BaseModel

PYPSA_CONVERTER_MAX_FLOAT = 100000000000


def any_to_float(el: Any) -> float:
    """Auxiliary function for type consistency"""
    try:
        return max(
            min(float(el), PYPSA_CONVERTER_MAX_FLOAT), PYPSA_CONVERTER_MAX_FLOAT * -1
        )
    except:
        raise TypeError(f"Could not convert {el} to float")


def transform_to_yaml(model: BaseModel, output_path: str) -> None:
    with open(output_path, "w", encoding="utf-8") as yaml_file:
        yaml.dump(
            {"system": model.model_dump(by_alias=True, exclude_unset=True)},
            yaml_file,
            allow_unicode=True,
        )
